#include "TROOT.h"
#include "TStyle.h"
#include "TFile.h"
#include "TTree.h"
#include "TCanvas.h"
#include "TTimer.h"
#include "TGFrame.h"
#include "TGTab.h"
#include "TBox.h"
#include "TFrame.h"
#include "TGStatusBar.h"
#include "TRootEmbeddedCanvas.h"
#include "TMath.h"
#include "RQ_OBJECT.h"
#include "TH1F.h"
#include "TH2F.h"
#include "TRandom.h"
#include "TUnixSystem.h"
#include "TApplication.h"
#include "sys/file.h"
#include "sys/time.h"
#include "libgen.h"
#include "stdint.h"
#include "fcntl.h"
#include "assert.h"

#include <iostream>
#include <sstream>

#include "ghost.h"
#include "ghost_site.h"

/* Calibration */

#include "calib/2016/run933.h"

/* Struct reader */
#include "ext_data_clnt.hh"
#include "ext_data_struct_info.hh"
extern "C" {
#include "ext_data_client.h"
#include "ext_h101.h"
}

/* Branch names
 * SIS3316: ADC*CH*MAXE* ADC*CH*T*
 * */
static const char* b_names[] =
{
/* Trigger logic branches */
	"MT", "TPAT", "TS_LO", "TS_HI", "TIMESTAMP",
	"TRIGT1", "TRIGT2", "TRIGT3", "TRIGT4",
	"TRIGT5", "TRIGT6", "TRIGT7", "TRIGT8",
	"TRIGT9", "TRIGT10", "TRIGT11", "TRIGT12",
/* LaBr Detector branches */
	"LABR1S", "LABR2S", "LABR3S", "LABR4S",
	"LABRL1S", "LABRL2S", "LABRL3S", "LABRL4S",
	"LABRL1T","LABRL2T","LABRL3T","LABRL4T",
	"ADC1CH1MAXE1", "ADC1CH5MAXE1", "ADC1CH9MAXE1", "ADC1CH13MAXE1",
	"ADC1CH1MAXE2", "ADC1CH5MAXE2", "ADC1CH9MAXE2", "ADC1CH13MAXE2",
/* HPGe Detector branches */
	"HPGE1S", "HPGE2S", "HPGE3S", "HPGE4S",
	"HPGEL1S", "HPGEL2S", "HPGEL3S", "HPGEL4S",
	"HPGEL1T","HPGEL2T","HPGEL3T","HPGEL4T",
	"ADC3CH1MAXE1", "ADC3CH5MAXE1", "ADC3CH9MAXE1", "ADC3CH13MAXE1",
	"ADC3CH1MAXE2", "ADC3CH5MAXE2", "ADC3CH9MAXE2", "ADC3CH13MAXE2",
	"CRM1", "CRM2", "CRM3", "CRM4", "CRM5", "CRM6", "CRM7", "CRM8",
/* Neutron detectors */
	"ADC2CH1GATE1", "ADC2CH5GATE1", "ADC2CH9GATE1", "ADC2CH13GATE1",
	"ADC2CH2GATE1", "ADC2CH6GATE1", "ADC2CH10GATE1", "ADC2CH14GATE1",
	"ADC2CH1GATE2", "ADC2CH5GATE2", "ADC2CH9GATE2", "ADC2CH13GATE2",
	"ADC2CH2GATE2", "ADC2CH6GATE2", "ADC2CH10GATE2", "ADC2CH14GATE2",
	"ADC2CH1GATE3", "ADC2CH5GATE3", "ADC2CH9GATE3", "ADC2CH13GATE3",
	"ADC2CH2GATE3", "ADC2CH6GATE3", "ADC2CH10GATE3", "ADC2CH14GATE3",
	"N1E", "N2E", "N3E", "N4E", "N5E", "N6E", "N7E", "N8E",
	"N1T", "N2T", "N3T", "N4T", "N5T", "N6T", "N7T", "N8T",
	"N1P", "N2P", "N3P", "N4P", "N5P", "N6P", "N7P", "N8P",
	"N1S", "N2S", "N3S", "N4S", "N5S", "N6S", "N7S", "N8S"
};

static const char* trlo_names[] = {
	"TRLORAW1","TRLORAW2","TRLORAW3","TRLORAW4",
	"TRLORAW5","TRLORAW6","TRLORAW7","TRLORAW8",
	"TRLORAW9","TRLORAW10","TRLORAW11","TRLORAW12",
	"TRLORAW13","TRLORAW14","TRLORAW15","TRLORAW16",
	"TRLOBDT1","TRLOBDT2","TRLOBDT3","TRLOBDT4",
	"TRLOBDT5","TRLOBDT6","TRLOBDT7","TRLOBDT8",
	"TRLOBDT9","TRLOBDT10","TRLOBDT11","TRLOBDT12",
	"TRLOBDT13","TRLOBDT14","TRLOBDT15","TRLOBDT16",
	"TRLOADT1","TRLOADT2","TRLOADT3","TRLOADT4",
	"TRLOADT5","TRLOADT6","TRLOADT7","TRLOADT8",
	"TRLOADT9","TRLOADT10","TRLOADT11","TRLOADT12",
	"TRLOADT13","TRLOADT14","TRLOADT15","TRLOADT16",
	"TRLOARD1","TRLOARD2","TRLOARD3","TRLOARD4",
	"TRLOARD5","TRLOARD6","TRLOARD7","TRLOARD8",
	"TRLOARD9","TRLOARD10","TRLOARD11","TRLOARD12",
	"TRLOARD13","TRLOARD14","TRLOARD15","TRLOARD16"
};

static const char* trlo_raw_names[] = {
	"LaBrOR L", "LaBrOR H", "HPGeOR L", "HPGeOR H",
	"LaBr M L", "LaBr M H", "HPGe M L", "HPGe M H",
	"  ---   ", "HPGe0deg", "RF      ", "Pulser  ",
	"  ---   ", "  ---   ", "  ---   ", "  ---   "
};

static const char* trlo_tpat_names[] = {
	"Singl LaBr H ", "Singl HPGe H ", "Coinc. L-L   ", "Coinc  L-H   ",
	"LaBr M H     ", "HPGe M H     ", "Zero Degree  ", "Pulser       ",
	"Singl LaBr  L", "Singl HPGe  L", "Coinc  H-L   ", "-            ",
	"-            ", "-            ", "-            ", "-            "
};

static const char* bc_names[] =
{
	"LABR1EC","LABR2EC","LABR3EC","LABR4EC",
	"HPGE1EC","HPGE2EC","HPGE3EC","HPGE4EC",
	"N1EC","N2EC","N3EC","N4EC",
	"N5EC","N6EC","N7EC","N8EC"
};

static const char* bcr_names[] =
{
	"LABR1ECR","LABR2ECR","LABR3ECR","LABR4ECR",
	"HPGE1ECR","HPGE2ECR","HPGE3ECR","HPGE4ECR",
	"N1ECR","N2ECR","N3ECR","N4ECR",
	"N5ECR","N6ECR","N7ECR","N8ECR"
};

/* Histogram names */
struct HistogramData
{
	const char *name;
	double bins;
	double min;
	double max;
	char log;
};

struct HistogramData h_data[] =
{
	{"Labr1", 4000, 0, 40000, 'n'},
	{"Labr2", 4000, 0, 40000, 'n'},
	{"Labr3", 4000, 0, 40000, 'n'},
	{"Labr4", 4000, 0, 40000, 'n'},
	{"Labr1_t_raw", 4000, 0, 4000, 'l'},
	{"Labr2_t_raw", 4000, 0, 4000, 'l'},
	{"Labr3_t_raw", 4000, 0, 4000, 'l'},
	{"Labr4_t_raw", 4000, 0, 4000, 'l'},
	{"Labr1_t / RF", 8000, -500, 500, 'l'},
	{"Labr2_t / RF", 8000, -500, 500, 'l'},
	{"Labr3_t / RF", 8000, -500, 500, 'l'},
	{"Labr4_t / RF", 8000, -500, 500, 'l'},
	{"HPGe1", 16000, 0, 9000000, 'n'},
	{"HPGe2", 16000, 0, 9000000, 'n'},
	{"HPGe3", 16000, 0, 9000000, 'n'},
	{"HPGe4", 16000, 0, 9000000, 'n'},
	{"HPGe1_t", 4000, -500, 500, 'n'},
	{"HPGe2_t", 4000, -500, 500, 'n'},
	{"HPGe3_t", 4000, -500, 500, 'n'},
	{"HPGe4_t", 4000, -500, 500, 'n'},
	{"HPGe1_t_raw", 4000, 0, 4000, 'l'},
	{"HPGe2_t_raw", 4000, 0, 4000, 'l'},
	{"HPGe3_t_raw", 4000, 0, 4000, 'l'},
	{"HPGe4_t_raw", 4000, 0, 4000, 'l'},
	{"HPGe1_t / RF", 4000, -500, 500, 'l'},
	{"HPGe2_t / RF", 4000, -500, 500, 'l'},
	{"HPGe3_t / RF", 4000, -500, 500, 'l'},
	{"HPGe4_t / RF", 4000, -500, 500, 'l'},
	{"LaBr_Sum", 2000, 0, 13000, 'n'},
	{"HPGe_Sum", 8000, 0, 12000, 'n'},
	{"Labr1_c", 5200, 0, 13000, 'n'},
	{"Labr2_c", 5200, 0, 13000, 'n'},
	{"Labr3_c", 5200, 0, 13000, 'n'},
	{"Labr4_c", 5200, 0, 13000, 'n'},
	{"HPGe1_c", 9600, 0, 12000, 'n'},
	{"HPGe2_c", 9600, 0, 12000, 'n'},
	{"HPGe3_c", 9600, 0, 12000, 'n'},
	{"HPGe4_c", 9600, 0, 12000, 'n'},
	{"Labr1_t_ref", 4000, -500, 500, 'l'},
	{"Labr2_t_ref", 4000, -500, 500, 'l'},
	{"Labr3_t_ref", 4000, -500, 500, 'l'},
	{"Labr4_t_ref", 4000, -500, 500, 'l'},
	{"HPGe1_t_ref", 4000, -500, 500, 'l'},
	{"HPGe2_t_ref", 4000, -500, 500, 'l'},
	{"HPGe3_t_ref", 4000, -500, 500, 'l'},
	{"HPGe4_t_ref", 4000, -500, 500, 'l'},
	{"Labr1 condition 1", 2000, 0, 10000, 'n'},
	{"Labr2 condition 1", 2000, 0, 10000, 'n'},
	{"Labr3 condition 1", 2000, 0, 10000, 'n'},
	{"Labr4 condition 1", 2000, 0, 10000, 'n'},
	{"Labr1 condition 2", 2000, 0, 10000, 'n'},
	{"Labr2 condition 2", 2000, 0, 10000, 'n'},
	{"Labr3 condition 2", 2000, 0, 10000, 'n'},
	{"Labr4 condition 2", 2000, 0, 10000, 'n'}
};

#define N_HISTOGRAMS sizeof(h_data) / sizeof(h_data[0])

struct HistogramData2D
{
	const char *name;
	double binsx;
	double minx;
	double maxx;
	double binsy;
	double miny;
	double maxy;
};

struct HistogramData2D h_data_2d[] =
{
	{"Coincidence HPGe-LaBr", 800, 0, 10000, 2000, 0, 10000},
	{"Coincidence LaBr-LaBr", 800, 0, 10000, 800, 0, 10000},
	{"HPGe1 vs. Time", 400, 0, 400, 1000, 0, 10000},
	{"HPGe2 vs. Time", 400, 0, 400, 1000, 0, 10000},
	{"HPGe3 vs. Time", 400, 0, 400, 1000, 0, 10000},
	{"HPGe4 vs. Time", 400, 0, 400, 1000, 0, 10000},
	{"HPGe1 vs. RF", 400, -400, 400, 1000, 0, 10000},
	{"HPGe2 vs. RF", 400, -400, 400, 1000, 0, 10000},
	{"HPGe3 vs. RF", 400, -400, 400, 1000, 0, 10000},
	{"HPGe4 vs. RF", 400, -400, 400, 1000, 0, 10000},
	{"HPGe1 over Time", 240, 0, 1e5, 3000, 0, 10000},
	{"HPGe2 over Time", 240, 0, 1e5, 3000, 0, 10000},
	{"HPGe3 over Time", 240, 0, 1e5, 3000, 0, 10000},
	{"HPGe4 over Time", 240, 0, 1e5, 3000, 0, 10000},
	{"LaBr1 vs. Time", 800, -100, 100, 400, 0, 10000},
	{"LaBr2 vs. Time", 800, -100, 100, 400, 0, 10000},
	{"LaBr3 vs. Time", 800, -100, 100, 400, 0, 10000},
	{"LaBr4 vs. Time", 800, -100, 100, 400, 0, 10000},
	{"LaBr1 vs. Rf", 800, -300, 300, 400, 0, 10000},
	{"LaBr2 vs. Rf", 800, -300, 300, 400, 0, 10000},
	{"LaBr3 vs. Rf", 800, -300, 300, 400, 0, 10000},
	{"LaBr4 vs. Rf", 800, -300, 300, 400, 0, 10000}
};

#define N_HISTOGRAMS_2D sizeof(h_data_2d) / sizeof(h_data_2d[0])

const char * f_names[] =
{
	"Raw E",
	"Raw T",
	"LaBr E",
	"LaBr E w/Cond",
	"HPGe E",
	"T-Ref",
	"T-Ref E",
	"T-RF",
	"T-RF E",
	"Sum/Coinc",
	"Timelines"
};

#define N_FRAMES sizeof(f_names) / sizeof(f_names[0])

/* Mapping from canvas ID to histogram index */
int n_canvases[N_FRAMES];
int canv2hist[N_FRAMES][32];
int canv2hist2d[N_FRAMES][32];

class myTH1F : public TH1F {
	public:
		myTH1F() : TH1F(nullptr, nullptr, 0, 0, 0), log(0) {}
		myTH1F(const char* name, Int_t nbinsx,
		    Double_t xlow, Double_t xup, char _log)
		    : TH1F(name, name, nbinsx, xlow, xup),log(_log) {}

		myTH1F(struct HistogramData data)
			: TH1F(data.name, data.name, data.bins, data.min,
			    data.max), log(data.log) {}

		void ResetHist() {this->Reset();} // *MENU*

		void ExecuteEvent(int ev, int x, int y) {
			(void) x;
			(void) y;
			if (ev == kKeyPress) {
				int key = gPad->GetEventX();
				if (key == 'r') {
					((myTH1F*)(gPad->
					    GetListOfPrimitives()->At(1)))->
					    ResetHist();
				}
			}
		}

		char log;

		//ClassDef(myTH1F,1);
};

class myTH2F : public TH2F {
	public:
		myTH2F(const char* name, Int_t nbinsx, Double_t xlow,
		    Double_t xup, Int_t nbinsy, Double_t ylow, Double_t yup)
		    : TH2F(name, name, nbinsx, xlow, xup, nbinsy, ylow, yup) {}

		myTH2F(struct HistogramData2D data) : TH2F(data.name, data.name,
		    data.binsx, data.minx, data.maxx, data.binsy, data.miny,
		    data.maxy) {}

		void ResetHist() {this->Reset();} // *MENU*

		//ClassDef(myTH2F,1);
};

class TreeReader : public TQObject {

	RQ_OBJECT("TreeReader");

	public:

	void struct_reader_init();
	void struct_read_loop(int);
	void struct_reader_close();
	void struct_reader_read();
	void struct_reader_get_values();
	void *run_struct_reader_loop(void);
	void *run_draw_loop(void);


	TreeReader(const char * _filename, const char * _options = 0)
	    : filename(_filename), is_struct_reader(0) {
		int showProgress = 1;

		memset(trlo_sc, 0, sizeof(trlo_sc));
		memset(trlo_sc_prev, 0, sizeof(trlo_sc_prev));
		memset(trlo_sc_sum, 0, sizeof(trlo_sc_sum));
		memset(h, 0, sizeof(h[0]) * N_HISTOGRAMS);
		memset(canv2hist, -1, sizeof(int) * N_FRAMES * 32);
		memset(canv2hist2d, -1, sizeof(int) * N_FRAMES * 32);

		createHistograms();
		fillCalibArrays();

		if(_options && strncmp(_options,"struct",6) == 0) {
			printf("Using struct reader as input\n");
			struct_reader_init();
			struct_read_loop(showProgress);
		}

		filename = _filename;
		TFile* file = new TFile(filename,"READ");
		f = file;
		ntuple = (TTree*)f->Get("h101");

		setBranchAddresses();

		first = 0;
		only_tv = 0;

		if(_options == 0 || strlen(_options) == 0) {
			struct timeval now, last_draw_time;

			gettimeofday(&last_draw_time, NULL);
			setupGui();
			styleHistograms();
			setLogScale();

			setupTimers();

			while(!gSystem->ProcessEvents()) {
				gettimeofday(&now, NULL);
				if (now.tv_sec > last_draw_time.tv_sec) {
					/*printf("Drawing...\n");*/
					refreshTree();
					draw();
					last_draw_time = now;
				}
			}
			gSystem->Abort();
		} else {
			int showScalers = 0;
			if(strncmp(_options,"tv",2) == 0) {
				printf("Creating tv spectra\n");
				only_tv = 1;
				printRealTime();
				treeLoop(showProgress);
				writeTv();
			} else if(strncmp(_options,"stat",4) == 0) {
				printf("Statistics\n");
				showScaler(showProgress,showScalers);
			} else if(strncmp(_options,"recal",5) == 0) {
				int n_peaks;
				int ret = sscanf(_options+5,"%d",&n_peaks);
				if (ret != 1) {
					printf("Must specify number of peaks "
					    "in 'recal' mode. %s (e.g.: recal "
					    "n)\n", _options);
					exit(1);
				}
				printf("Recalibration with %d peaks\n",n_peaks);
				recalibrateTree(showProgress,n_peaks);
			} else {
				if(strncmp(_options,"cal",3) != 0) {
					printf("Unknown option %s\n",_options);
					return;
				}
				extendTree(showProgress);
			}
			gSystem->Abort();
		}
	}

	~TreeReader() {
		f->Close();
		delete f;
		f = 0;
		frame->Cleanup();
		delete frame;
	}

	void setupTimers() {
		t1 = new TTimer();
		t1->Start(500);
		t2 = new TTimer();
		t2->Start(500);
	}

	void createHistograms() {

		memset(h,0,sizeof(h));

		TStyle* myStyle = gROOT->GetStyle("Modern");
		myStyle->SetOptStat("neimr");
		myStyle->SetStatH(0.3);
		myStyle->SetStatW(0.2);
		myStyle->cd();

		/* RAW_E */
		for (int i = 0; i < N_LABR; ++i) {
			h[Labr1 + i] = new myTH1F(h_data[Labr1 + i]);
			canv2hist[RAW_E][i + 1] = Labr1 + i;
		}
		for (int i = 0; i < N_HPGE; ++i) {
			h[HPGe1 + i] = new myTH1F(h_data[HPGe1 + i]);
			canv2hist[RAW_E][i + N_LABR + 1] = HPGe1 + i;
		}
		/* RAW_T */
		for (int i = 0; i < N_LABR; ++i) {
			h[Labr1_t_raw + i] =
			    new myTH1F(h_data[Labr1_t_raw + i]);
			canv2hist[RAW_T][i + 1] = Labr1_t_raw + i;
		}
		for (int i = 0; i < N_HPGE; ++i) {
			h[HPGe1_t_raw + i] =
			    new myTH1F(h_data[HPGe1_t_raw + i]);
			canv2hist[RAW_T][i + N_LABR + 1] = HPGe1_t_raw + i;
		}
		/* LABR_E */
		for(int i = 0; i < N_LABR; ++i) {
			h[Labr1_c + i] = new myTH1F(h_data[Labr1_c + i]);
			canv2hist[LABR_E][i + 1] = Labr1_c + i;
		}
		/* HPGE_E */
		for(int i = 0; i < N_HPGE; ++i) {
			h[HPGe1_c + i] = new myTH1F(h_data[HPGe1_c + i]);
			canv2hist[HPGE_E][i + 1] = HPGe1_c + i;
		}
		/* T vs. Ref */
		for (int i = 0; i < N_LABR; ++i) {
			h[Labr1_t_ref + i] =
			    new myTH1F(h_data[Labr1_t_ref + i]);
			canv2hist[T_VS_REF][i + 1] = Labr1_t_ref + i;
		}
		for (int i = 0; i < N_HPGE; ++i) {
			h[HPGe1_t_ref + i] =
			    new myTH1F(h_data[HPGe1_t_ref + i]);
			canv2hist[T_VS_REF][i + N_LABR + 1] = HPGe1_t_ref + i;
		}
		/* T vs. RF */
		for (int i = 0; i < N_LABR; ++i) {
			h[Labr1_t_RF + i] =
			    new myTH1F(h_data[Labr1_t_RF + i]);
			canv2hist[T_VS_RF][i + 1] = Labr1_t_RF + i;
			/* w1 */
			box_t_vs_rf_labr[i] = new TBox(0,0,1,1);
			box_t_vs_rf_labr[i]->SetFillColor(kOrange);
			box_t_vs_rf_labr[i]->SetLineColor(kOrange);
			box_t_vs_rf_labr[i]->SetFillStyle(3002);
			/* w2 */
			box_t_vs_rf_labr[i+N_LABR] = new TBox(0,0,1,1);
			box_t_vs_rf_labr[i+N_LABR]->SetFillColor(kGreen);
			box_t_vs_rf_labr[i+N_LABR]->SetLineColor(kGreen);
			box_t_vs_rf_labr[i+N_LABR]->SetFillStyle(3002);
		}
		for (int i = 0; i < N_HPGE; ++i) {
			h[HPGe1_t_RF + i] =
			    new myTH1F(h_data[HPGe1_t_RF + i]);
			canv2hist[T_VS_RF][i + N_LABR + 1] = HPGe1_t_RF + i;
			/* w1 */
			box_t_vs_rf_hpge[i] = new TBox(0,0,1,1);
			box_t_vs_rf_hpge[i]->SetFillColor(kOrange);
			box_t_vs_rf_hpge[i]->SetLineColor(kOrange);
			box_t_vs_rf_hpge[i]->SetFillStyle(3002);
		}
		/* SUM_COINC */
		// Sum
		h[LaBr_Sum] = new myTH1F(h_data[LaBr_Sum]);
		canv2hist[SUM_COINC][1] = LaBr_Sum;
		h[HPGe_Sum] = new myTH1F(h_data[HPGe_Sum]);
		canv2hist[SUM_COINC][2] = HPGe_Sum;
		// Coincidence
		h2[Coincidence_LaBr_LaBr] =
		    new myTH2F(h_data_2d[Coincidence_LaBr_LaBr]);
		canv2hist2d[SUM_COINC][3] = Coincidence_LaBr_LaBr;
		h2[Coincidence_HPGe_LaBr] =
		    new myTH2F(h_data_2d[Coincidence_HPGe_LaBr]);
		canv2hist2d[SUM_COINC][4] = Coincidence_HPGe_LaBr;
		/* LABR_E_COND */
		for (int i = 0; i < N_LABR; ++i) {
			// LaBr energy spectra with applied condition
			h[Labr1condition1 + i] =
			    new myTH1F(h_data[Labr1condition1 + i]);
			canv2hist[LABR_E_COND][i + 1] = Labr1condition1 + i;
			h[Labr1condition2 + i] =
			    new myTH1F(h_data[Labr1condition2 + i]);
			canv2hist[LABR_E_COND][N_LABR + i + 1] =
				Labr1condition2 + i;
		}

		for(int i = 0; i < N_HPGE; ++i) {
			h[HPGe1_t + i] = new myTH1F(h_data[HPGe1_t + i]);
			// HPGe time-energy spectra
			h2[HPGe1vsTime + i] =
			    new myTH2F(h_data_2d[HPGe1vsTime + i]);
			canv2hist2d[T_VS_REF_E][N_LABR + i + 1] =
			    HPGe1vsTime + i;
			// HPGe time vs RF spectra
			h2[HPGe1vsRf + i] =
			    new myTH2F(h_data_2d[HPGe1vsRf + i]);
			canv2hist2d[T_VS_RF_E][N_LABR + i + 1] =
			    HPGe1vsRf + i;
			// HPGe energy spectra over time (to check for drifts)
			h2[HPGe1overTime + i] =
			    new myTH2F(h_data_2d[HPGe1overTime + i]);
			canv2hist2d[TIMELINES][i + 1] = HPGe1overTime + i;
		}
		for (int i = 0; i < N_LABR; ++i) {
			h2[LaBr1vsTime + i] =
			    new myTH2F(h_data_2d[LaBr1vsTime + i]);
			canv2hist2d[T_VS_REF_E][i + 1] = LaBr1vsTime + i;
			h2[LaBr1vsRf + i] =
			    new myTH2F(h_data_2d[LaBr1vsRf + i]);
			canv2hist2d[T_VS_RF_E][i + 1] = LaBr1vsRf + i;
		}
	}

	void setLogScale() {
#if ELOG
		TVirtualPad* vp;
		for(int i = 0; i < N_LABR; ++i) {
			vp = ec[LABR_E]->GetCanvas()->cd(i+1);
			vp->SetLogy();
			vp = ec[4]->GetCanvas()->cd(i+1);
			vp->SetLogy();
			vp = ec[7]->GetCanvas()->cd(i+1);
		}
		for (int i = 0; i < N_HPGE; ++i) {
			vp = ec[1]->GetCanvas()->cd(i+1);
			vp->SetLogy();
			vp = ec[1]->GetCanvas()->cd(4+i+1);
			vp = ec[7]->GetCanvas()->cd(8+i+1);
		}
#endif
	}

	void styleHistograms() {
		size_t i;
		int j;
		double x = 0.02;
		double y = 0.02;

		PRINTD("%s BEGIN\n", __func__);
		for (i = 0; i < N_FRAMES; ++i) {
			PRINTD("frame %lu, n_canvases = %d\n",
			    i, n_canvases[i]);
			if (n_canvases[i] > 8) {
				x = 0.04;
				y = 0.04;
			}
			else if (n_canvases[i] > 4) {
				x = 0.07;
				y = 0.07;
			} else {
				x = 0.04;
				y = 0.04;
			}
			for (j = 1; j <= n_canvases[i]; ++j) {
				PRINTD("  canvas %d -> hist %d, hist2d %d\n",
				    j, canv2hist[i][j], canv2hist2d[i][j]);
				if (canv2hist[i][j] >= 0) {
					h[canv2hist[i][j]]->
						GetXaxis()->SetLabelSize(x);
					h[canv2hist[i][j]]->
						GetYaxis()->SetLabelSize(y);
				}
				if (canv2hist2d[i][j] >= 0) {
					h2[canv2hist2d[i][j]]->
						GetXaxis()->SetLabelSize(x);
					h2[canv2hist2d[i][j]]->
						GetYaxis()->SetLabelSize(y);
				}
			}
		}
		PRINTD("%s END\n", __func__);
	}

	void createEmbeddedCanvas(enum FrameName a_f, size_t nx, size_t ny)
	{
		size_t i;
		TCanvas *c;

		ec[a_f] = new TRootEmbeddedCanvas(f_names[a_f], cf[a_f], 1200,
		    900);
		c = ec[a_f]->GetCanvas();
		c->Divide(nx, ny);
		for (i = 0; i < nx * ny; ++i) {
			c->cd(i+1)->SetBit(TPad::kCannotMove);
			//c->cd(i+1)->AddExec("ec0_1_exec","onKeyPress()");
			if(h_data[canv2hist[a_f][i+1]].log == 'l') {
				c->cd(i+1)->SetLogy();
			}
		}
		n_canvases[a_f] = nx * ny;
	}

	void setupGui()
	{
		size_t i;
		frame = new TGMainFrame(gClient->GetRoot(),400,400);

		tab = new TGTab(frame,1200,900);
		for (i = 0; i < N_FRAMES; ++i) {
			cf[i] = tab->AddTab(f_names[i]);
		}

		status = new TGStatusBar(frame,50,10,kHorizontalFrame);
		status->SetParts(3);

		createEmbeddedCanvas(RAW_E, 2, 4);
		createEmbeddedCanvas(RAW_T, 2, 4);
		createEmbeddedCanvas(LABR_E, 2, 2);
		createEmbeddedCanvas(LABR_E_COND, 2, 4);
		createEmbeddedCanvas(HPGE_E, 2, 2);
		createEmbeddedCanvas(T_VS_REF, 2, 4);
		createEmbeddedCanvas(T_VS_REF_E, 2, 4);
		createEmbeddedCanvas(T_VS_RF, 2, 4);
		createEmbeddedCanvas(T_VS_RF_E, 2, 4);
		createEmbeddedCanvas(SUM_COINC, 2, 2);
		createEmbeddedCanvas(TIMELINES, 2, 2);

		for (i = 0; i < N_FRAMES; ++i) {
			cf[i]->AddFrame(ec[i],
			    new TGLayoutHints(kLHintsExpandX | kLHintsExpandY,
				10, 10, 10, 1));
			cf[i]->Resize(cf[i]->GetDefaultSize());
		}

		frame->AddFrame(tab, new TGLayoutHints(kLHintsExpandX |
		    kLHintsExpandY, 10, 10, 10, 1));
		frame->AddFrame(status, new TGLayoutHints(kLHintsBottom |
		    kLHintsLeft | kLHintsExpandX, 0, 0, 2, 0));

		frame->MapSubwindows();
		frame->Resize(frame->GetDefaultSize());
		frame->MapWindow();

		frame->SetWindowName("GHOST - Gamma3 Histogramming and Online "
		    "Spectra Tool");
	}

	void setBranchAddresses() {
		for(int i = 0; i < N_LABR; ++i) {
			/* LABR */
			b_labr_h_e[i] = 0;
			b_labr_l_e[i] = 0;
			b_labr_sis_e[i] = 0;
			b_labr_sis_b[i] = 0;
			b_labr_s[i] = 0;
			labr_s_prev[i] = 0;
			ntuple->SetBranchAddress(b_names[LABRL1T + i],
			    &labrl_t[i], &b_labrl_t[i]);
			ntuple->SetBranchAddress(b_names[ADC1CH1MAXE1 + i],
			    &labr_sis_b[i], &b_labr_sis_b[i]);
			ntuple->SetBranchAddress(b_names[ADC1CH1MAXE2 + i],
			    &labr_sis_e[i], &b_labr_sis_e[i]);
			ntuple->SetBranchAddress(b_names[LABR1S + i],
			    &labr_s[i], &b_labr_s[i]);
			ntuple->SetBranchAddress(b_names[LABRL1S + i],
			    &labrl_s[i], &b_labrl_s[i]);
		}
		for(int i = 0; i < N_HPGE; ++i) {
			/* HPGE */
			b_hpge_e[i] = 0;
			b_hpge_b[i] = 0;
			b_hpge_s[i] = 0;
			b_hpge_pileup[i] = 0;
			hpge_s_prev[i] = 0;
			ntuple->SetBranchAddress(b_names[ADC3CH1MAXE2 + i],
			    &hpge_e[i], &b_hpge_e[i]);
			ntuple->SetBranchAddress(b_names[ADC3CH1MAXE1 + i],
			    &hpge_b[i], &b_hpge_b[i]);
			ntuple->SetBranchAddress(b_names[HPGEL1T + i],
			    &hpgel_t[i], &b_hpgel_t[i]);
			ntuple->SetBranchAddress(b_names[HPGE1S + i],
			    &hpge_s[i], &b_hpge_s[i]);
			ntuple->SetBranchAddress(b_names[HPGEL1S + i],
			    &hpgel_s[i], &b_hpgel_s[i]);
		}
		for (int i = 0; i < N_TRIG; ++i) {
			ntuple->SetBranchAddress(b_names[TRIGT1 + i],
			    &trig_t[i], &b_trig_t[i]);
		}
		for (int i = 0; i < N_CRM; ++i) {
			b_hpge_crm[i] = 0;
			hpge_crm_prev[i] = 0;
			ntuple->SetBranchAddress(b_names[CRM1 + i],
			    &hpge_crm[i], &b_hpge_crm[i]);
		}
		ntuple->SetBranchAddress(b_names[TPAT], &tpat, &b_tpat);
		ntuple->SetBranchAddress(b_names[TS_LO], &time_lo, &b_time_lo);
		ntuple->SetBranchAddress(b_names[TS_HI], &time_hi, &b_time_hi);

		for(int i = 0; i < N_TRLO; ++i) {
			b_trlo_sc[TRLORAW1 + i] = 0;
			b_trlo_sc[TRLOBDT1 + i] = 0;
			b_trlo_sc[TRLOADT1 + i] = 0;
			b_trlo_sc[TRLOARD1 + i] = 0;
			ntuple->SetBranchAddress(trlo_names[TRLORAW1 + i],
			    &trlo_sc[i].raw, &b_trlo_sc[i]);
			ntuple->SetBranchAddress(trlo_names[TRLOBDT1 + i],
			    &trlo_sc[i].bdt, &b_trlo_sc[TRLOBDT1 + i]);
			ntuple->SetBranchAddress(trlo_names[TRLOADT1 + i],
			    &trlo_sc[i].adt, &b_trlo_sc[TRLOADT1 + i]);
			ntuple->SetBranchAddress(trlo_names[TRLOARD1 + i],
			    &trlo_sc[i].ard, &b_trlo_sc[TRLOARD1 + i]);
		}
	}

#ifdef USE_STEPWISE_CALIB
	void calibrate_stepwise(int n) {
		for(int i = 0; i < N_LABR; i++) {
			double labr_h =
			    (double)(labr_h_e[i]) + r.Rndm() - 0.5;
			double labr_sis =
			    (double)(labr_sis_e[i]) + r.Rndm() - 0.5;
			double labr_l =
			    (double)(labr_l_e[i]) + r.Rndm() - 0.5;

			/* High range LaBr */
			int idx = i + (n * 7);
			labr_h_e_calib[i] = (labr_h * labr_h * labr_h *
					cubic_labr_stepwise[idx]) +
				(labr_h * labr_h *
				 	quadratic_labr_stepwise[idx]) +
				(labr_h * slopes_labr_stepwise[idx]) +
					offsets_labr_stepwise[idx];
			labr_sis_e_calib[i] = (labr_sis * labr_sis * labr_sis *
				       cubic_labr_stepwise[idx]) +
				(labr_sis * labr_sis *
				 	quadratic_labr_stepwise[idx]) +
				(labr_sis * slopes_labr_stepwise[idx])
					+ offsets_labr_stepwise[idx];

			/* Low range LaBr */
			labr_l_e_calib[i] = labr_l * slopes_labr_l[i]
				+ offsets_labr_l[i];
		}
		for(int i = 0; i < N_HPGE; i++) {
			double hpge =
			    (double)(hpge_e[i] - hpge_b[i]) + r.Rndm() - 0.5;
			double hpge_bin =
			    (double)(h_data[HPGe1 + i].bins) /
			    (double)(h_data[HPGe1 + i].max);

			/* Struck HPGe */
			hpge_e_calib[i] = hpge * hpge * hpge * cubic_hpge[i] *
				  hpge_bin * hpge_bin * hpge_bin
				+ hpge * hpge * quadratic_hpge[i] *
				  hpge_bin * hpge_bin
				+ hpge * slopes_hpge[i] * hpge_bin
				+ offsets_hpge[i];

			/* HPGe Baseline correct */
			hpge_e_nobl[i] = hpge_e[i] - hpge_b[i];
		}
	}
#endif

	void calibrate() {
		for(int i = 0; i < N_LABR; i++) {
			double labr_sis = (double)(labr_sis_e[i]
			    - labr_sis_b[i]) + r.Rndm() - 0.5;
			double hpge = (double)(hpge_e[i] - hpge_b[i])
			    + r.Rndm() - 0.5;
			double hpge_bin =
			    (double)(h_data[HPGe1 + i].bins) /
			    (double)(h_data[HPGe1 + i].max);
			double labr_bin =
			    (double)(h_data[Labr1 + i].bins) /
			    (double)(h_data[Labr1 + i].max);
			double labr_sis_bin = labr_sis * labr_bin;

			labr_sis_e_calib[i] = (labr_sis_bin * labr_sis_bin
				* labr_sis_bin * cubic_labr_h[i])
				+ (labr_sis_bin * labr_sis_bin
					* quadratic_labr_h[i])
				+ (labr_sis_bin * slopes_labr_h[i])
				+ offsets_labr_h[i];

			hpge_e_calib[i] = hpge * hpge * hpge * cubic_hpge[i] *
				  hpge_bin * hpge_bin * hpge_bin
				+ hpge * hpge * quadratic_hpge[i] * hpge_bin * 
				  hpge_bin
				+ hpge * slopes_hpge[i] * hpge_bin
				+ offsets_hpge[i];
			hpge_e_nobl[i] = hpge_e[i] - hpge_b[i];

			PRINTD("labr_e[%d] = %.2f\n", i, labr_sis_e_calib[i]);
			
			//double labr_h = (double)(labr_h_e[i]) + r.Rndm() - 0.5;
			//double labr_l = (double)(labr_l_e[i]) + r.Rndm() - 0.5;
			/*labr_h_e_calib[i] = (labr_h * labr_h * labr_h *
				  cubic_labr_h[i])
				+ (labr_h * labr_h * quadratic_labr_h[i])
				+ (labr_h * slopes_labr_h[i])
				+ offsets_labr_h[i];*/
			/*labr_l_e_calib[i] = labr_l * slopes_labr_l[i]
				+ offsets_labr_l[i];*/
		}
	}

	void refreshTree() {
		int tries = 4;
		int keys_readable = 0;
		struct flock fl;
		int fd = f->GetFd();
		int ret;

		fl.l_whence = SEEK_SET;
		fl.l_start = 0;
		fl.l_len = 1;

		/*printf("Refreshing tree...\n");*/

		while(tries-- && (keys_readable == 0)) {
			/* Get file lock */
			fl.l_type = F_RDLCK;
			ret = fcntl(fd, F_SETLKW, &fl);
			if (ret == -1) {
				if (errno == EINTR) {
					continue;
				} else {
					perror("fcntl(F_SETLKW, F_RDLCK");
				}
			}

			/*printf("Got file lock.\n");*/

			if(ntuple->GetDirectory()->ReadKeys() != 0)
			    keys_readable = 1;

			usleep(100);
		}
		if(tries == 0) {
			printf("Tried 10 times. Resetting...");
			// Reset, if we cannot find our keys
			f->Close();
			delete f;
			f = 0;
			ntuple = 0;
			first = 0;
			exit(1);
		}

		/*printf("Refreshing");*/
		ntuple->Refresh();

		/* Release lock */
		fl.l_type = F_UNLCK;
		ret = fcntl(fd, F_SETLK, &fl);
		if (ret == -1) {
			perror("fcntl(F_SETLK, F_UNLCK");
		}

		if(first == ntuple->GetEntries()) {
			return;
		}
	}

	void extendTree(int showProgress = 0) {
		int nof_entries = ntuple->GetEntries() - first;
		int status_update_counter = 0;

		char friendname[255];
		char filebase[1000];
		char buf[255];

		strcpy(filebase, filename);
		filebase[strlen(filebase)-5] = '\0';
		sprintf(friendname,"%s_calib.root", filebase);
		printf("friendname = %s\n", friendname);

		TFile* cal = new TFile(friendname,"recreate");
		TTree* h102 = new TTree("h102","SYNC level tree");

#if 0
		for(int i = 0; i < 8; ++i) {
			char buf[255];
			sprintf(buf,"%s/F",bc_names[i]);
			h102->Branch(bc_names[i],&labr_h_e_calib[i],buf);
			if(i < 4) {
				sprintf(buf,"%s/F",bc_names[8+i]);
				h102->Branch(bc_names[8+i],&hpge_e_calib[i],buf);  
			}
		}
#endif
		for(int i = 0; i < N_LABR; ++i) {
			sprintf(buf,"%s/F",bc_names[LABR1EC + i]);
			h102->Branch(bc_names[LABR1EC + i],
			    &labr_sis_e_calib[i], buf);
		}
		for(int i = 0; i < N_HPGE; ++i) {
			sprintf(buf,"%s/F",bc_names[HPGE1EC + i]);
			h102->Branch(bc_names[HPGE1EC + i],
			    &hpge_e_calib[i],buf);
		}

#ifdef USE_STEPWISE_CALIB
		int labr_last_calibration = (sizeof(labr_timestamp_start) /
		    sizeof(labr_timestamp_start[0])) - 1;
		int labr_current_calibration = 0;
#endif 

		// Activate only necessary branches
		ntuple->SetBranchStatus("*",0);
		ntuple->SetBranchStatus(b_names[TIMESTAMP], 1);
		for(int i = 0; i < N_LABR; ++i) {
			ntuple->SetBranchStatus(b_names[ADC1CH1MAXE1 + i], 1);
			ntuple->SetBranchStatus(b_names[ADC1CH1MAXE2 + i], 1);
		}
		for(int i = 0; i < N_HPGE; ++i) {
			ntuple->SetBranchStatus(b_names[ADC3CH1MAXE1 + i], 1);
			ntuple->SetBranchStatus(b_names[ADC3CH1MAXE2 + i], 1);
		}

		// Loop over entries
		for(int entry = first; entry < ntuple->GetEntries(); ++entry) {
			ntuple->GetEntry(entry);

			/* Build 64 bit timestamp */
			timestamp = ((uint64_t)time_hi) << 32 | time_lo;

			if(entry == first) ts_first = timestamp;

			if(showProgress == 1 &&
			    ++status_update_counter == 100000) {
				double progress = ((double)(entry-first)) /
				    ((double)nof_entries);
				fprintf(stderr,"Filling histograms: " GREEN
				    "%8d" BLACK "(" GREEN
				    "%.1f%%" BLACK ")\r", entry + 1,
				    progress * 100);
				fflush(stderr);
				status_update_counter = 0;
			}
			if(entry == ntuple->GetEntries()-1) {
				fprintf(stderr,"\n");
			}

			PRINTD("%d\n",entry);

#ifdef USE_STEPWISE_CALIB
			// Set current calibration
			if(labr_current_calibration < labr_last_calibration
			    && timestamp >
			    labr_timestamp_start[labr_current_calibration+1])
			{
				++labr_current_calibration;
				printf("\nChanging labr calibration to %d "
				    "starting at timestamp %d\n",
				    labr_current_calibration, timestamp);
			}
			calibrate_stepwise(labr_current_calibration);
#else
			calibrate();
#endif

			h102->Fill();
		}

		h102->Write();
		cal->Close();
	}

	typedef struct recalibration_set_s {
		int det;
		int from;
		int to;
		int t_from;
		int t_to;
		int n_peaks;
		float offset;
		float slope;
	} recalibration_set;

	void recalibrateTree(int showProgress = 0, int n_peaks = 2) {
		int nof_entries = ntuple->GetEntries() - first;
		int status_update_counter = 0;

		char friendname[255];
		char recalibration_filename[255];
		char filebase[1000];

		strcpy(filebase, filename);
		filebase[strlen(filebase)-5] = '\0';
		sprintf(friendname,"%s_recalib.root", filebase);
		printf("friendname = %s\n",friendname);

		sprintf(recalibration_filename,"%s_recalib.dat", filebase);
		printf("recalibration = %s\n",recalibration_filename);
		printf("commencing recalibration with %d peaks.\n",n_peaks);

		TFile* cal = new TFile(friendname, "recreate");
		TTree* tree = new TTree("h103",
		    "SYNC level tree (recalibrated HPGE)");

		// Read recalibration file
		FILE* recalibration_file = 0;
		char line[255];

		recalibration_file = fopen(recalibration_filename, "r");
		if(recalibration_file == 0) {
			printf("Error opening recalibration file: %s\n",
			    recalibration_filename);
			exit(1);
		}

		// Pre-allocate 100 records and 4 detectors
		recalibration_set recal_data[4][100];
		size_t n_records[4];
		for (int i = 0; i < 4; ++i) { n_records[i] = 0;}

		// Go through the lines
		while(fgets(line,sizeof(line),recalibration_file) != 0) {
			int det = 0;
			int from = 0, to = 0;
			int t_from = 0, t_to = 0;
			float* heights;
			float* energies;
			float* widths;
			float offset = 0;
			float slope = 0;
			//float dummy;
			char* str = 0;

			// Allocate memory for the data 
			heights = (float*) malloc(n_peaks * sizeof(float));
			energies = (float*) malloc(n_peaks * sizeof(float));
			widths = (float*) malloc(n_peaks * sizeof(float));

			// Parse line
			str = strtok(line," ");
			if(str != 0) sscanf(str,"%d",&det);
			str = strtok(0," ");
			if(str != 0) sscanf(str,"%d",&from);
			str = strtok(0," ");
			if(str != 0) sscanf(str,"%d",&to);
			str = strtok(0," ");
			if(str != 0) sscanf(str,"%d",&t_from);
			str = strtok(0," ");
			if(str != 0) sscanf(str,"%d",&t_to);

			for(int i = 0; i < n_peaks; ++i) {
				str = strtok(0," ");
				if(str != 0) sscanf(str,"%f",&heights[i]);
				str = strtok(0," ");
				if(str != 0) sscanf(str,"%f",&energies[i]);
				str = strtok(0," ");
				if(str != 0) sscanf(str,"%f",&widths[i]);
			}

			str = strtok(0," ");
			if(str != 0) sscanf(str,"%f",&offset);
			str = strtok(0," ");
			if(str != 0) sscanf(str,"%f",&slope);

			// Test output
			printf("%lu %d %d %d %f %f\n", n_records[det-1], det,
			    from, to, offset, slope);
			for(int i = 0; i < n_peaks; ++i) {
				printf("  %f %f %f\n", energies[i], heights[i],
				    widths[i]);
			}
			fflush(stdout);

			// Store data
			recal_data[det-1][n_records[det-1]].det = det;
			recal_data[det-1][n_records[det-1]].from = from;
			recal_data[det-1][n_records[det-1]].to = to;
			recal_data[det-1][n_records[det-1]].t_from = t_from;
			recal_data[det-1][n_records[det-1]].t_to = t_to;
			recal_data[det-1][n_records[det-1]].offset = offset;
			recal_data[det-1][n_records[det-1]].slope = slope;

			free(heights);
			free(widths);
			free(energies);
			++n_records[det-1];
		}
		fclose(recalibration_file);

		// Create new branches
		char buf[255];
		for(int i = 0; i < N_LABR; ++i) {
			sprintf(buf,"%s/F", bcr_names[LABR1ECR + i]);
			tree->Branch(bcr_names[LABR1ECR + i],
			    &labr_sis_e_calib[i],buf);
		}
		for(int i = 0; i < N_HPGE; ++i) {
			sprintf(buf,"%s/F",bcr_names[HPGE1ECR + i]);
			tree->Branch(bcr_names[HPGE1ECR + i],
			    &hpge_e_calib[i],buf);
		}

		// Activate only necessary branches
		ntuple->SetBranchStatus("*",0);
		ntuple->SetBranchStatus(b_names[TIMESTAMP], 1);
		for(int i = 0; i < N_LABR; ++i) {
			ntuple->SetBranchStatus(b_names[ADC1CH1MAXE1 + i], 1);
			ntuple->SetBranchStatus(b_names[ADC1CH1MAXE2 + i], 1);
		}
		for(int i = 0; i < N_HPGE; ++i) {
			ntuple->SetBranchStatus(b_names[ADC3CH1MAXE1 + i], 1);
			ntuple->SetBranchStatus(b_names[ADC3CH1MAXE2 + i], 1);
		}

		// Loop over tree
		int cal_idx = 0;
		for(int entry = first; entry < ntuple->GetEntries(); ++entry) {

			// Update recal data
			while(entry >= recal_data[0][cal_idx].to && cal_idx < (int)(n_records[0])-1) {
				++cal_idx;
				printf("Using calibration data from %d to %d "
				    "(offset=%.2f, slope=%.5f)\n",
				    recal_data[0][cal_idx].from, recal_data[0][cal_idx].to,
				    recal_data[0][cal_idx].offset, recal_data[0][cal_idx].slope);
			}

			ntuple->GetEntry(entry);

			/* Build 64 bit timestamp */
			timestamp = ((uint64_t)time_hi) << 32 | time_lo;
			if(entry == first) ts_first = timestamp;

			// Status update
			if(showProgress == 1 && ++status_update_counter == 100000) {
				double progress = ((double)(entry-first)) /
				    ((double)nof_entries);
				fprintf(stderr,"Filling histograms: " GREEN 
					"%8d" BLACK "(" GREEN
					"%.1f%%" BLACK ")\r",
					entry+1,progress*100);
				fflush(stderr);
				status_update_counter = 0;
			}
			if(entry == ntuple->GetEntries()-1) { fprintf(stderr,"\n"); }

			PRINTD("%d\n",entry);

			calibrate(); // Do the normal calibration

			// Recalibrate
			for(int i = 0; i < N_LABR; ++i) {
				hpge_e_calib[i] = (hpge_e_calib[i]
				    * recal_data[i][cal_idx].slope)
				    + recal_data[i][cal_idx].offset;
			}


			tree->Fill();

		}

		tree->Write();
		cal->Close();
	}

	void showScaler(int showProgress = 0, int showScalers = 0) {
		// Manual loop
		int nof_entries = ntuple->GetEntries() - first;
		int status_update_counter = 0;
		char run_nr[8];
		char buf[256];
		char path[256];
		char inputname[1000];
		char* path_ptr = path;
		char* filename_ptr = inputname;

		Int_t entries = ntuple->GetEntries();

		/* Open output file */
		FILE* sc_out = 0;

		strcpy(path, filename);
		strcpy(inputname, filename);

		path_ptr = dirname(path);
		filename_ptr = basename(inputname);

		PRINTD("%s, path_ptr: %s, file_ptr: %s\n", filename, path_ptr,
		    filename_ptr); fflush(stdout);
		PRINTD("%s, path: %s, file: %s\n", filename, path, filename);
		fflush(stdout);

		sprintf(buf,"%s",path_ptr);
		gSystem->MakeDirectory(buf);

		if(strcmp(filename_ptr, "stream.root") == 0) {
			sprintf(buf, "%s/stream_scalers.dat", path_ptr);
		} else {
			if(strncmp(filename_ptr, "run", 3) == 0) {
				strncpy(run_nr, (filename_ptr + 3),3);
				run_nr[3] = '\0';
				sprintf(buf, "%s/run%s_scalers.dat", path,
				    run_nr);
			} else {
				printf("No stream file, and no run number "
				    "given. Saving tv spectra as unknown.\n");
				sprintf(buf, "%s/unknown_scalers.dat", path);
			}
		}
		PRINTD("Opening scaler output file: %s\n", buf);
		sc_out = fopen(buf, "w");
		if(!sc_out) {
			printf("Could not open scaler output file '%s'\n.",
			    buf);
		}

		// Activate only necessary branches
		ntuple->SetBranchStatus("*",0);
		ntuple->SetBranchStatus(b_names[TIMESTAMP], 1);
		for(int i = 0; i < N_LABR; ++i) {
			ntuple->SetBranchStatus(b_names[LABR1S + i], 1);
			ntuple->SetBranchStatus(b_names[LABRL1S + i], 1);
		}
		for(int i = 0; i < N_HPGE; ++i) {
			ntuple->SetBranchStatus(b_names[HPGE1S + i], 1);
			ntuple->SetBranchStatus(b_names[HPGEL1S + i], 1);
		}
		for(int i = 0; i < N_TRLO; ++i) {
			ntuple->SetBranchStatus(trlo_names[TRLORAW1 + i], 1);
			ntuple->SetBranchStatus(trlo_names[TRLOBDT1 + i], 1);
			ntuple->SetBranchStatus(trlo_names[TRLOADT1 + i], 1);
			ntuple->SetBranchStatus(trlo_names[TRLOARD1 + i], 1);
		}

		/* Loop through scaler events */
		fprintf(sc_out,"# scaler1_raw scaler1_before_deadtime "
		    "scaler1_after_deadtime scaler1_after_reduction ...\n");
		for(int entry = first; entry < entries; ++entry) {

			if(showProgress == 1 &&
			    ++status_update_counter == 100000) {
				double progress = ((double)(entry-first)) /
				    ((double)nof_entries);
				fprintf(stderr,"Computing statistics: " GREEN
				    "%8d" BLACK "(" GREEN "%.1f%%" BLACK ")\r",
				    entry + 1, progress * 100);
				fflush(stdout);
				status_update_counter = 0;
			}
			if(entry == entries-1) { fprintf(stderr,"\n"); }

			PRINTD("%d\n",entry);

			b_trlo_sc[0]->GetEntry(entry);

			if(trlo_sc[0].raw == 0) continue;
			ntuple->GetEntry(entry);

			/* Increase scalers */
			for(int i = 0; i < N_TRLO; ++i) {
				if (entry != 0) {
					trlo_sc_sum[i].raw += trlo_sc[i].raw
					    - trlo_sc_prev[i].raw;
					trlo_sc_sum[i].bdt += trlo_sc[i].bdt
					    - trlo_sc_prev[i].bdt;
					trlo_sc_sum[i].adt += trlo_sc[i].adt
					    - trlo_sc_prev[i].adt;
					trlo_sc_sum[i].ard += trlo_sc[i].ard
					    - trlo_sc_prev[i].ard;
				}
				trlo_sc_prev[i].raw = trlo_sc[i].raw;
				trlo_sc_prev[i].bdt = trlo_sc[i].bdt;
				trlo_sc_prev[i].adt = trlo_sc[i].adt;
				trlo_sc_prev[i].ard = trlo_sc[i].ard;
			}

			// Scalers update
			if(labr_s[0] != 0) {
				updateScalerRatesWithWrap();
			}

			/* Print scaler event */
			if(showScalers == 1) {
				printf("Trlo Scalers ......................."
				    "......................................."
				    ".......\n");
				printf("Name               RAW #            "
				      "	B.DT      A.DT      A.RD Deadtime   "
				      "\n");
				for(int i = 0; i < N_TRLO; ++i) {
					printf("%2d: %s %9d # %2d:%s %9d %9d "
					    "%9d %.2f\n", i + 1,
					    trlo_raw_names[i], trlo_sc[i].raw,
					    i + 1, trlo_tpat_names[i],
					    trlo_sc[i].bdt, trlo_sc[i].adt,
					    trlo_sc[i].ard, 100 * (double)
					    (trlo_sc[i].adt) / (double)
					    (trlo_sc[i].bdt));
				}
				printf("\n");
			}

			/* Print scaler event to file */
			for(int i = 0; i < N_TRLO; ++i) {
				fprintf(sc_out,"%u %u %u %u ",
				    trlo_sc[i].raw, trlo_sc[i].bdt,
				    trlo_sc[i].adt, trlo_sc[i].ard);
			}
			for(int i = 0; i < N_LABR; ++i) {
				fprintf(sc_out,"%d %d ", labrl_rate[i],
				    labr_rate[i]);
			}
			for(int i = 0; i < N_HPGE; ++i) {
				fprintf(sc_out,"%d %d ", hpgel_rate[i],
				    hpge_rate[i]);
			}
			for(int i = 0; i < N_CRM; ++i) {
				fprintf(sc_out,"%d ", hpge_crm_rate[i]);
			}
			fprintf(sc_out,"\n");
		}

		/* Print summary */
		printRealTime();
		getRunNumber(run_nr);
		printf("Trlo Scalers for run %s ............................."
		    "........................................................"
		    ".........\n", run_nr);
		printf("Name                        RAW #                       "
		    "         B.DT               A.DT               A.RD Deadtime Reduction "
		    "\n");
		for(int i = 0; i < N_TRLO; ++i) {
			printf("%2d: %s %18lu # %2d:%s %18lu %18lu %18lu "
			"%5.2f %5.2f\n", i + 1, trlo_raw_names[i],
			    trlo_sc_sum[i].raw, i + 1,
			    trlo_tpat_names[i],
			    trlo_sc_sum[i].bdt,
			    trlo_sc_sum[i].adt,
			    trlo_sc_sum[i].ard,
			    100 * (1. -
				  (double)(trlo_sc_sum[i].adt)
				/ (double)(trlo_sc_sum[i].bdt)),
			    (double)(trlo_sc_sum[i].adt) /
			    (double)(trlo_sc_sum[i].ard)
			    );
		}
	}


	void fillLaBrEnergies(int i)
	{
		PRINTD("%s BEGIN\n", __func__);
		assert(h[Labr1 + i] != NULL);

		h[Labr1 + i]->Fill(labr_e);

		if(only_tv == 1) return;

		float time_l = labrl_t[i] / 8.;

		/* Calibrated LaBr energies */
		if((   (tpat & TPAT_LABR_SINGLE_H) > 0
		    || (tpat & TPAT_LABR_SINGLE_L) > 0
		    || (tpat & TPAT_HPGE_SINGLE_L) > 0
		    || (tpat & TPAT_COINCIDENCE_LL) > 0
		    )
		    && time_l > 0)
		{
			float tdiff = time_l - t_rf;

			assert(h[Labr1_c + i] != NULL);
			h[Labr1_c + i]->Fill(labr_sis_e_calib[i]);

			// Condition 1
			if(time_l > 0 && t_rf > 0
			    && tdiff > labr_w1[2*i]
			    && tdiff < labr_w1[(2*i)+1])
			    //&& tdiff > labr_t_min[i]
			    //&& tdiff < labr_t_max[i])
			{
				assert(h[Labr1condition1 + i] != NULL);
				h[Labr1condition1 + i]->Fill(
				    labr_sis_e_calib[i]);
			}
			// Condition 2
			if(time_l > 0 && t_rf > 0
			    && tdiff > labr_w2[(2*i)]
			    && tdiff < labr_w2[(2*i)+1])
			    //&& tdiff > labr_t_min[i]
			    //&& tdiff < labr_t_max[i])
			{
				assert(h[Labr1condition2 + i] != NULL);
				h[Labr1condition2 + i]->Fill(
				  labr_sis_e_calib[i]);
			}
		}
		PRINTD("%s END\n", __func__);
	}

	void fillLaBrTimes(int i)
	{
		int all_others = 0;
		float t_trig = trig_t[TRIG_LABR_OR_L] / 8.;
		float time_l = labrl_t[i] / 8.;
		float tdiff = time_l - t_trig;

		(void)tdiff;

		PRINTD("%s BEGIN\n", __func__);
		if(time_l <= 0) return;

		h[Labr1_t_raw + i]->Fill(labrl_t[i]);

#if 1
		for(int j = 0; j < N_LABR; ++j) {
			if(j!=i && labrl_t[j] > 0) {
				all_others = 1;
			}
		}
		for(int j = 0; j < N_HPGE; ++j) {
			if(hpgel_t[j] > 0) {
				all_others = 1;
			}
		}
#endif

		if (0) { printf("%d", all_others); }

		if(t_ref > 0) {
			all_others = 0;
			for(int j = 0; j < N_LABR; ++j) {
				if(j != ref && j != i && labrl_t[j] > 0) {
					all_others = 1;
				}
			}
			for(int j = 0; j < N_HPGE; ++j) {
				if(hpgel_t[j] > 0) {
					all_others = 1;
				}
			}

			if(all_others == 0) {
				assert(h[Labr1_t_ref + i] != NULL);
				h[Labr1_t_ref + i]->Fill(time_l - t_ref);
				h2[LaBr1vsTime + i]->Fill(time_l - t_ref,
				    labr_sis_e_calib[i]);
			}
		}
		if(all_others == 0 && t_rf > 0) {
			assert(h2[LaBr1vsRf + i] != NULL);
			h2[LaBr1vsRf + i]->Fill(time_l - t_rf,
			    labr_sis_e_calib[i]);
		}
		PRINTD("%s END\n", __func__);
	}

	void fillLaBrIndividualTimes(int i)
	{
		float time_l = labrl_t[i] / 8.;
		PRINTD("%s (i = %d) BEGIN\n", __func__, i);
#if 0
		if(labrl_t[i]>0 && labrl_t[0]>0) {
			/* LABR vs. Labr 1 HIGH */
			h[Labr1_t_RF + i]->Fill(time_l - t_ref);
		}
#else
		if(time_l > 0 && t_rf > 0
		    && labr_sis_e_calib[i]>1000. /* LABR vs RF */
		    //&& (time_l - t_rf) > labr_t_min[i]
		    //&& (time_l - r_rf) < labr_t_max[i])
		) {
			assert(h[Labr1_t_RF + i] != NULL);
			h[Labr1_t_RF + i]->Fill(time_l - t_rf);
		}
#endif
		PRINTD("%s END\n", __func__);
	}

	void fillLaBrHistograms()
	{
		PRINTD("%s BEGIN\n", __func__);
		for(int i = 0; i < N_LABR; ++i) {
			labr_e = labr_sis_e[i] - labr_sis_b[i];

			if((USED_CONDITION) && labr_e > thr_labr_sis[i]) {
				fillLaBrEnergies(i);
				fillLaBrTimes(i);
				fillLaBrIndividualTimes(i);
			}
		}
		PRINTD("%s END\n", __func__);
	}

	void fillHPGeTimes(int i)
	{
		float time_l = hpgel_t[i] / 8.;
		float t_trig = trig_t[TRIG_HPGE_OR_L] / 8.;
		float t_diff = time_l - t_trig;
		float t_diff_l = time_l - t_trig;
		float t_diff_ref = time_l - t_ref;
		float t_diff_rf = time_l - t_rf;

		(void)t_diff;
		(void)t_diff_l;
		PRINTD("%s BEGIN\n", __func__);

		/* Raw low time */
		h[HPGe1_t_raw + i]->Fill(hpgel_t[i]);

#if 1
		// Raw time spectra
		int all_others = 0;
		for(int j = 0; j < N_HPGE; ++j) {
			if(j != i && hpgel_t[j] > 0) {
				all_others = 1;
			}
		}
		for(int j = 0; j < N_LABR; ++j) {
			if(labrl_t[j] > 0) {
				all_others = 1;
			}
		}

		if(0) printf("%d", all_others);
#endif

		/* Self-trigger LOW */
		//if(t_trig > 0)
		//	h[HPGe1_t_raw + i]->Fill(t_diff_l);

		/* vs. Reference time */
		if(t_ref > 0 && time_l > 0) {
			assert(h[HPGe1_t_ref + i] != NULL);
			h[HPGe1_t_ref + i]->Fill(t_diff_ref);
		}

		/* vs. RF time */
		if(all_others == 0 && t_rf > 0 && time_l > 0) {
			assert(h[HPGe1_t_RF + i] != NULL);
			h[HPGe1_t_RF + i]->Fill(t_diff_rf);
		}

		// Individual time spectra (2D)
		if(all_others == 0 && t_rf > 0 && time_l > 0){
			assert(h2[HPGe1vsRf + i] != NULL);
			h2[HPGe1vsRf + i]->Fill(t_diff_rf, hpge_e_calib[i]);
		}
		if(t_ref > 0 && time_l > 0){
			assert(h2[HPGe1vsTime + i] != NULL);
			h2[HPGe1vsTime + i]->Fill(t_diff_ref, hpge_e_calib[i]);
		}

		PRINTD("%s END\n", __func__);
	}

	void fillHPGeEnergies(int i)
	{
		(void) i;
		PRINTD("%s BEGIN\n", __func__);
		assert(h[HPGe1 + i] != NULL);
		h[HPGe1 + i]->Fill(hpge_e_nobl[i]);

		if(only_tv == 1) return;

		float time_l = hpgel_t[i] / 8.;

		// Cut on HPGE OR trigger, a valid time
		// in this channel, and no pileup
		if(/*(tpat & TPAT_HPGE_SINGLE_L) > 0 &&*/ time_l > 0
			/*&& hpge_pileup[i] == 0*/
			/*&& t_diff > hpge_t_min[i]
			  && t_diff < hpge_t_max[i])*/
		) {
			assert(h[HPGe1_c + i] != NULL);
			assert(h2[HPGe1overTime + i] != NULL);

			h[HPGe1_c + i]->Fill(hpge_e_calib[i]);
			h2[HPGe1overTime + i]->Fill(
			    (timestamp - timestamp_begin) / 1e8,
			    hpge_e_calib[i]);
			PRINTD("time = %.2f\n",
			    (timestamp - timestamp_begin) / 1e8);
		}
		PRINTD("%s END\n", __func__);
	}

	void fillHPGeIndividualTimes(int i)
	{
		PRINTD("%s BEGIN\n", __func__);
#if 0
		float t_diff_rf = time_l - t_rf;

		/*if(hpgel_t[i] > 0 && labrl_t[0] > 0 ) */
		if(hpgel_t[i] > 0 && trig_t[TRIG_HPGE_OR_L] > 0
			&& hpge_e_calib[i] > 1000.) {
			//h[HPGe1_t + i]->Fill(t_diff_ref);
			assert(h[HPGe1_t + i] != NULL);
			h[HPGe1_t + i]->Fill(t_diff_rf);
		}
#else
		(void) i;
#endif
		PRINTD("%s END\n", __func__);
	}

	void fillHPGeHistograms()
	{
		PRINTD("%s BEGIN\n", __func__);
		for(int i = 0; i < N_HPGE; ++i) {
			if(hpge_e_nobl[i] > thr_hpge[i]) {
				fillHPGeEnergies(i);
				fillHPGeTimes(i);
				fillHPGeIndividualTimes(i);
			}
		}
		PRINTD("%s END\n", __func__);
	}

	void fillSumHistograms()
	{
		PRINTD("%s BEGIN\n", __func__);
		// LaBr sum
		for(int i = 0; i < N_LABR; ++i) {
			if(labr_sis_e[i] - labr_sis_b[i] > thr_labr_sis[i]
			    && labr_sis_e_calib[i] > 100) {
				assert(h[LaBr_Sum] != NULL);
				h[LaBr_Sum]->Fill(labr_sis_e_calib[i]);
			}
		}
		// HPGe sum
		for(int i = 0; i < N_HPGE; ++i) {
			if((hpge_e_nobl[i] > thr_hpge[i]) && hpgel_t[i] > 0
			    /*&& hpge_pileup[i] == 0*/)
			{
				assert(h[HPGe_Sum] != NULL);
				h[HPGe_Sum]->Fill(hpge_e_calib[i]);
			}
		}
		PRINTD("%s END\n", __func__);
	}

	void fillCoincidenceMatrices()
	{
		assert(h2[Coincidence_HPGe_LaBr] != NULL);
		assert(h2[Coincidence_LaBr_LaBr] != NULL);

		PRINTD("%s BEGIN\n", __func__);

		// Coincidence matrix HPGe LaBr
		for(int i = 0; i < N_LABR; ++i ) {
			float t_labr = labrl_t[i] / 8.;
			float t_labr_rf = t_labr - t_rf;
			if(labr_sis_e[i] - labr_sis_b[i] < thr_labr_sis[i])
				continue; 
			if((tpat & TPAT_COINCIDENCE_LL)) {
				for(int j = 0; j < N_HPGE; ++j) {
					if(hpge_e_nobl[j] < thr_hpge[j])
						continue;
					float t_hpge = hpgel_t[j] / 8.;
					float t_hpge_rf = t_hpge - t_rf;


					/* Use coincidence windows */
#if 1
					if((t_labr_rf > labr_w3[2*i]
					    && t_labr_rf < labr_w3[(2*i)+1])
					|| (t_labr_rf > labr_w1[2*i]
					    && t_labr_rf < labr_w1[(2*i)+1])
					|| (t_hpge_rf > hpge_w1[2*j]
					    && t_hpge_rf < hpge_w1[(2*j)+1])
					    )
					    //|| (labr_h_e_calib[i] < 150)
					    /*&& hpge_e_calib[j] > 1000)*/
#else
						(void)t_hpge_rf;
						(void)t_labr_rf;
#endif
					{
						PRINTD("  HPGE-LABR.\n");
						h2[Coincidence_HPGe_LaBr]
						    ->Fill(labr_sis_e_calib[i],
						    hpge_e_calib[j]);
					}
				}
			}
			// Labr-Labr
			if((tpat & TPAT_LABR_M_LO) || (tpat & TPAT_LABR_M)) {
				for(int j = i+1; j < 4; ++j) {
					float t_labr2 = labrl_t[j]/8.;
					float t_labr2_rf = t_labr2 - t_rf;
					if(labr_sis_e[j] - labr_sis_b[j]
					    < thr_labr_sis[j]) continue; 
					//    || labr_h_e_calib[i] < 550*/) continue;
#if 1
					if ((t_labr_rf > labr_w1[2*i])
					    && (t_labr_rf < labr_w1[(2*i)+1])
					    && (t_labr2_rf > labr_w1[2*j])
					    && (t_labr2_rf < labr_w1[(2*j)+1]))
					    //||
					    //(labr_h_e_calib[i] < 150)
					    //&& labr_h_e_calib[j] > 1500)
#else
						(void)t_labr2_rf;
#endif
					{
						PRINTD("  LABR-LABR.\n");
						h2[Coincidence_LaBr_LaBr]
						->Fill(labr_sis_e_calib[i],
						    labr_sis_e_calib[j]);
						h2[Coincidence_LaBr_LaBr]
						->Fill(labr_sis_e_calib[j],
						    labr_sis_e_calib[i]);
					}
				}
			}
		}

		PRINTD("%s END\n", __func__);
	}

	void update_timestamp() {
		timestamp = ((uint64_t)time_hi) << 32 | time_lo;
	}

	void treeLoop(int showProgress = 0, int gui = 0)
	{
		PRINTD("%s BEGIN\n", __func__);
		// Manual loop
		int nof_entries = ntuple->GetEntries() - first;
		int status_update_counter = 0;
		int statusbar_update_counter = 0;

		timestamp = 0;

		if(only_tv == 1) {
			// Activate only necessary branches
			ntuple->SetBranchStatus("*",0);
			ntuple->SetBranchStatus(b_names[TS_LO], 1);
			ntuple->SetBranchStatus(b_names[TS_HI], 1);
			for(int i = 0; i < N_LABR; ++i) {
				ntuple->SetBranchStatus(
				    b_names[ADC1CH1MAXE1 + i], 1);
				ntuple->SetBranchStatus(
				    b_names[ADC1CH1MAXE2 + i], 1);
			}
			for(int i = 0; i < N_HPGE; ++i) {
				ntuple->SetBranchStatus(
				    b_names[ADC3CH1MAXE1 + i], 1);
				ntuple->SetBranchStatus(
				    b_names[ADC3CH1MAXE2 + i], 1);
			}
			ntuple->SetBranchStatus(b_names[TPAT], 1);
		}

#ifdef USE_STEPWISE_CALIB
		int labr_last_calibration = (sizeof(labr_timestamp_start) /
		    sizeof(labr_timestamp_start[0])) - 1;
		int labr_current_calibration = 0;
#endif

		int cnt = 0;
		printf("Filling %d entries.\n", nof_entries);

		for(int entry = first; entry < ntuple->GetEntries(); ++entry) {
			ntuple->GetEntry(entry);

			PRINTD("Entry %d\n", entry);

			update_timestamp();

			if(entry == first) ts_first = timestamp;

			if(showProgress == 0 && first == 0 &&
			    (++status_update_counter == 1000)) {
				status_update_counter = 0;
				double progress = (double)(entry-first) /
				    (double)(nof_entries);
				printf("\rCatching up: " GREEN "%.1f%%" BLACK
				    "                   ", 100 * progress);
				fflush(stdout);
			}

			/* Allow to have interaction while filling histograms */
			cnt++;
			if(first != 0 && cnt == 50) {
				gSystem->ProcessEvents();
				cnt = 0;
			}
			if(only_tv == 0 && first == 0 && cnt == 5000) {
				redrawHistograms();
				gSystem->ProcessEvents();
				cnt = 0;
			}
			if(first == 0 && entry == first) {
				timestamp_begin = timestamp;
			}

#ifdef USE_STEPWISE_CALIB
			// Write tv spectra and clear histograms
			if(labr_current_calibration < labr_last_calibration
			    && timestamp >
			    labr_timestamp_start[labr_current_calibration+1]) 
			{
				++labr_current_calibration;
				printf("\nChanging labr calibration to %d "
				    "starting at timestamp %d\n",
				    labr_current_calibration,timestamp);
				writeTv(labr_current_calibration + 1);
				clearHistograms();
			}
#endif

			if(gui == 1 && ++statusbar_update_counter == 10000) {
				char buf[255];
				statusbar_update_counter = 0;
				if(first == 0) {
					double progress =
					    (double)(entry-first) /
					    (double)(nof_entries);
					sprintf(buf,"Catching up: %.1f      ",
					    100*progress);
				} else {
					sprintf(buf,"Events: %lld",
					    ntuple->GetEntries());
				}
				status->SetText(buf, 0);
				gPad->Modified();
				gPad->Update();
				gSystem->ProcessEvents();
			}

			if(showProgress == 1 &&
			    ++status_update_counter == 100000)
			{
				double progress = ((double)(entry-first)) /
				    ((double)nof_entries);
				fprintf(stderr,
				    "Filling histograms: " GREEN "%8d" BLACK
				    "(" GREEN "%.1f%%" BLACK ")\r",
				    entry + 1, progress * 100);
				fflush(stderr);
				status_update_counter = 0;
			}
			if(entry == ntuple->GetEntries()-1) {
				fprintf(stderr,"\n");
			}

			PRINTD("%d\n", entry);

#ifdef USE_STEPWISE_CALIB
			calibrate_stepwise(labr_current_calibration);
#else
			calibrate();
#endif

			update_reference_values();

			fillHistograms();
		}

		if (only_tv == 0) {
			updateScalerRates();
		}
	}

#define REF 0
	void update_reference_values() {
		/* Time of RF pulse (beam pickup) */
		t_rf = trig_t[TRIG_RF] / 8.;
		/* Time of Reference detector */
		t_ref = labrl_t[REF] / 8.;
	}

	void updateScalerRates() {
		for (int i = 0; i < N_LABR; ++i) {
			labr_rate[i] = labr_s_cur[i] - labr_s_prev[i];
			labr_s_prev[i] = labr_s_cur[i];
			labrl_rate[i] =	labrl_s_cur[i] - labrl_s_prev[i];
			labrl_s_prev[i] = labrl_s_cur[i];
		}
		for (int i = 0; i < N_HPGE; ++i) {
			hpge_rate[i] = hpge_s_cur[i] - hpge_s_prev[i];
			hpge_s_prev[i] = hpge_s_cur[i];
			hpgel_rate[i] =	hpgel_s_cur[i] - hpgel_s_prev[i];
			hpgel_s_prev[i] = hpgel_s_cur[i];
		}
		for (int i = 0; i < N_CRM; ++i) {
			hpge_crm_rate[i] = hpge_crm_cur[i] - hpge_crm_prev[i];
			hpge_crm_prev[i] = hpge_crm_cur[i];
		}

	}

	void updateScalerValues() {
		if(labr_s[0] != 0 || labrl_s[0] != 0) {
			for(int i = 0; i < N_LABR; ++i) {
				labr_s_cur[i] = labr_s[i];
				labrl_s_cur[i] = labrl_s[i];
			}
			for(int i = 0; i < N_HPGE; ++i) {
				hpge_s_cur[i] = hpge_s[i];
				hpgel_s_cur[i] = hpgel_s[i];
			}
			for(int i = 0; i < N_CRM; ++i) {
				hpge_crm_cur[i] = hpge_crm[i];
			}
		}
	}

#define MAX_SCALER_26BIT 0x4000000
	void updateScalerRatesWithWrap()
	{	
		for(int i = 0; i < N_LABR; ++i) {
			if (labr_s[i] > labr_s_prev[i]) {
				labr_rate[i] = labr_s[i] - labr_s_prev[i];
			} else {
				labr_rate[i] = (labr_s[i] + MAX_SCALER_26BIT)
				    - labr_s_prev[i];
			}
			if (labrl_s[i] > labrl_s_prev[i]) {
				labrl_rate[i] = labrl_s[i] - labrl_s_prev[i];
			} else {
				labrl_rate[i] = (labrl_s[i] + MAX_SCALER_26BIT)
				    - labrl_s_prev[i];
			}
		}
		for(int i = 0; i < N_HPGE; ++i) {
			if (hpge_s[i] > hpge_s_prev[i]) {
				hpge_rate[i] = hpge_s[i]-hpge_s_prev[i];
			} else {
				hpge_rate[i] = (hpge_s[i] + MAX_SCALER_26BIT)
				    - hpge_s_prev[i];
			}
			if (hpgel_s[i] > hpgel_s_prev[i]) {
				hpgel_rate[i] = hpgel_s[i] - hpgel_s_prev[i];
			} else {
				hpgel_rate[i] = (hpgel_s[i] + MAX_SCALER_26BIT)
				    - hpgel_s_prev[i];
			}
		}
		for(int i = 0; i < N_CRM; ++i) {
			if (hpge_crm[i] > hpge_crm_prev[i]) {
				hpge_crm_rate[i] = hpge_crm[i] -
				    hpge_crm_prev[i];
			} else {
				hpge_crm_rate[i] = (hpge_crm[i]
				    + MAX_SCALER_26BIT)	- hpge_crm_prev[i];
			}
		}

		for (int i = 0; i < N_LABR; ++i) {
			labr_s_prev[i] = labr_s[i];
			labrl_s_prev[i] = labrl_s[i];
		}
		for (int i = 0; i < N_HPGE; ++i) {
			hpge_s_prev[i] = hpge_s[i];
			hpgel_s_prev[i] = hpgel_s[i];
		}
		for (int i = 0; i < N_CRM; ++i) {
			hpge_crm_prev[i] = hpge_crm[i];
		}
	}

	void fillHistograms() {
		fillLaBrHistograms();
		fillHPGeHistograms();

		if(only_tv == 0) {
			fillSumHistograms();
			fillCoincidenceMatrices();
			updateScalerValues();
		}

	}

	void printRealTime(int gui = 0) {
		float realtime;

		// Print real time
		ntuple->GetEntry(0);
		start = ((uint64_t)time_hi) << 32 | time_lo;
		ntuple->GetEntry(ntuple->GetEntries()-1);
		stop = ((uint64_t)time_hi) << 32 | time_lo;

		realtime = (stop - start) / 1e8;

		if(gui == 1) {
			char buf[255];
			sprintf(buf,"Real time: %.2f s", realtime);
			status->SetText(buf,1);
		}
		printf("Real time: %.2f s.\n", realtime);
	}

	void draw() {
		int showProgress = 0;
		int gui = 1;
		uint64_t ts = 0;

		treeLoop(showProgress, gui);

		printRealTime(gui);

		// Print scalers
		printf(  "--------- Scalers ----------- 1 sec -------------\n");
		printf(  " Name   | Lo Thr, Rate      | Hi Thr, Rate\n");
		for(int i = 0; i < N_LABR; ++i) {
			printf(" LaBr %d | %11d %6d   %11d %6d       \n",
			    i+1, labrl_s_cur[i], labrl_rate[i],
			    labr_s_cur[i], labr_rate[i]);
		}
		for(int i = 0; i < N_HPGE; ++i) {
			printf(" HPGe %d | %11d %6d   %11d %6d CRM: %11d %6d\n",
			    i+1, hpgel_s_cur[i], hpgel_rate[i],
			    hpge_s_cur[i], hpge_rate[i],
			    hpge_crm_cur[i], hpge_crm_rate[i]);
		}

		// Scale rate histogram
		ts_last = ts;
		int tdiff = ts_last - ts_first;
		if(tdiff <= 0) tdiff = 1;

		gSystem->ProcessEvents();

		first = (Int_t)ntuple->GetEntries();

		/*printf("Redrawing histograms ... \n");*/
		redrawHistograms();

		gSystem->ProcessEvents();

		/*printf("Writing tv files...\n");*/
		writeTv();

	}

	void redrawHistogramsOnFrame(enum FrameName a_f)
	{
		for (int i = 0; i < n_canvases[a_f]; ++i) {
			assert(ec[a_f] != NULL);

			ec[a_f]->GetCanvas()->cd(i + 1);

			/* a 1D histogram */
			if (canv2hist[a_f][i+1] >= 0) {
				assert(h[canv2hist[a_f][i+1]] != NULL);
				h[canv2hist[a_f][i + 1]]->Draw();
			}
			/* a 2D histogram */
			if (canv2hist2d[a_f][i+1] >= 0) {
				assert(h2[canv2hist2d[a_f][i+1]] != NULL);
				h2[canv2hist2d[a_f][i + 1]]->Draw("col");
			}
			if (canv2hist2d[a_f][i+1] >= 0
			    && canv2hist[a_f][i+1] >= 0) {
				printf("Error: 1d and 2d histogram in the same"
				    " pad (a_f = %d, i = %d)", a_f, i);
				exit(1);
			}
			gSystem->ProcessEvents();
		}
	}

	void redrawHistograms()
	{
		size_t i;
		size_t n;
		TCanvas* c[N_FRAMES];

		n = tab->GetCurrent();

		for (i = 0; i < N_FRAMES; ++i) {
			c[i] = ec[i]->GetCanvas();
		}

		/* Redraw special histograms */
#if 0
		if (n == T_VS_RF) {
			for (i = 0; i < N_HPGE; ++i) {
				assert(h2[HPGe1vsTime + i] != NULL);
				c[T_VS_RF]->cd(N_LABR+i+1);
				h2[HPGe1vsTime + i]->Draw("col");
				gSystem->ProcessEvents();
			}
		}
		if (n == TIMELINES) {
			for (i = 0; i < N_HPGE; ++i) {
				assert(h2[HPGe1overTime + i] != NULL);
				c[TIMELINES]->cd(i+1);
				h2[HPGe1overTime + i]->Draw("col");
				gSystem->ProcessEvents();
			}
		}
#endif

		/* Redraw normal 1D histograms */
		redrawHistogramsOnFrame((enum FrameName)n);

		/* Annotate histograms */
		if (n == T_VS_RF) {
			/* Add time windows to ToF plots */
			for (i = 0; i < N_LABR; ++i) {
				TPad *pad;
				double ymax;
				pad = (TPad *)c[T_VS_RF]->cd(i + 1);
				ymax = pad->GetUymax();
				if (pad->GetLogy()) {
					ymax = TMath::Power(10, ymax);
				}
				box_t_vs_rf_labr[i]->DrawBox(labr_w1[i*2],
				    pad->GetUymin(), labr_w1[i*2+1], ymax);
				box_t_vs_rf_labr[i+N_LABR]->DrawBox(
				    labr_w3[i*2], pad->GetUymin(),
				    labr_w3[i*2+1], ymax);
				gSystem->ProcessEvents();
			}
			for (i = 0; i < N_HPGE; ++i) {
				TPad *pad;
				double ymax;
				pad = (TPad *)c[T_VS_RF]->cd(i + 1 + N_LABR);
				ymax = pad->GetUymax();
				if (pad->GetLogy()) {
					ymax = TMath::Power(10, ymax);
				}
				box_t_vs_rf_hpge[i]->DrawBox(hpge_w1[i*2],
				    pad->GetUymin(), hpge_w1[i*2+1], ymax);
				/*box_t_vs_rf_hpge[i+N_LABR]->DrawBox(
				    hpge_w2[i*2], pad->GetUymin(),
				    hpge_w2[i*2+1], ymax);*/
				gSystem->ProcessEvents();
			}
		}
		c[n]->Update();
		gSystem->ProcessEvents();
	}

	void getRunNumber(char* run_nr) {
		char inputname[1000];
		char* filename_ptr = inputname;

		strcpy(inputname, filename);
		filename_ptr = basename(inputname);
		strncpy(run_nr, (filename_ptr + 3), 3);
		run_nr[3] = '\0';
	}

	void fillCalibArrays() {
		for(int i = 0; i < N_LABR; ++i) {
			calib_offsets[i] = offsets_labr_h[i];
			calib_slopes[i] = slopes_labr_h[i];
			calib_quad[i] = quadratic_labr_h[i];
			calib_cubic[i] = cubic_labr_h[i];
		}
		for(int i = 0; i < N_HPGE; ++i) {
			calib_offsets[i+N_LABR] = offsets_hpge[i];
			calib_slopes[i+N_LABR] = slopes_hpge[i];
		}
	}

	void clearHistograms() {
		// List of histograms
		int h_map[] = {
			Labr1, Labr2, Labr3, Labr4,
			HPGe1, HPGe2, HPGe3, HPGe4
		};
		for(int i = 0; i < (int)(sizeof(h_map)/sizeof(int)); ++i) {
			h[h_map[i]]->Reset();
		}
	}

	void writeTv(int index = 0) {
		// Write TV spectra to disk (with an appended 'index',
		// if index != 0)
		FILE* tv_out;
		// List of histograms
		int h_map[] = {
			Labr1, Labr2, Labr3, Labr4,
			HPGe1, HPGe2, HPGe3, HPGe4
		};

		for(int i = 0; i < (int)(sizeof(h_map)/sizeof(int)); ++i) {
			char buf[256];
			char calbuf[256];
			char path[256];
			char run_nr[8];
			char inputname[1000];
			const char* path_ptr = path;
			const char* filename_ptr = inputname;

			if (is_struct_reader) {
				path_ptr = "/tmp";
				filename_ptr = "stream";
			} else {
				strcpy(path, filename);
				strcpy(inputname, filename);
				path_ptr = dirname(path);
				filename_ptr = basename(inputname);
			}


			PRINTD("%s, path_ptr: %s, file_ptr: %s\n",
			    filename,path_ptr,filename_ptr);
			fflush(stdout);
			PRINTD("%s, path: %s, file: %s\n", filename, path,
			    filename);
			fflush(stdout);

			sprintf(buf,"%s/tv",path_ptr);
			sprintf(calbuf,"%s/tv",path_ptr);
			gSystem->MakeDirectory(buf);

			if(strncmp(filename_ptr,"stream",6) == 0) {
				sprintf(buf,"%s/tv/%s_%s.tv", path_ptr,
				    filename_ptr, h_data[h_map[i]].name);
				sprintf(calbuf,"%s/tv/%s_%s.cal", path_ptr,
				    filename_ptr, h_data[h_map[i]].name);
			} else {
				if(strncmp(filename_ptr,"run",3) == 0) {
					strncpy(run_nr,(filename_ptr + 3),3);
					run_nr[3] = '\0';
					if(index == 0) {
						sprintf(buf,"%s/tv/run%s_%s.tv",
						    path,run_nr,
						    h_data[h_map[i]].name);
						sprintf(calbuf,"%s/tv/run%s_%s.cal",
						    path,run_nr,
						    h_data[h_map[i]].name);
					} else {
						sprintf(buf,"%s/tv/run%s_%s_%d.tv",
						    path,run_nr,
						    h_data[h_map[i]].name,
						    index);
						sprintf(calbuf,"%s/tv/run%s_%s_%d.cal",
						    path,run_nr,
						    h_data[h_map[i]].name,
						    index);
					}
				} else {
					printf("No stream file, and no run number given. "
					    "Saving tv spectra as unknown.\n");
					sprintf(buf,"%s/tv/unknown_%s.tv",path,
					    h_data[h_map[i]].name);
					sprintf(calbuf,"%s/tv/unknown_%s.cal",
					    path, h_data[h_map[i]].name);
				}
			}
			PRINTD("Opening tv output file: %s\n",buf);
			tv_out = fopen(buf,"w");
			if(!tv_out) {
				printf("Could not open tv output file '%s'\n.",buf);
				continue;
			}

			fprintf(tv_out,"# Real time: %lu s.\n", stop - start);
			for(int n = 0; n < h[h_map[i]]->GetNbinsX(); ++n) {
				fprintf(tv_out,"%d\n",(int)h[h_map[i]]->GetBinContent(n));
			}
			fclose(tv_out);

			PRINTD("Opening calibration output file: %s\n",calbuf);
			tv_out = fopen(calbuf,"w");
			if(!tv_out) {
				printf("Could not open tv calibration output file '%s'\n.",calbuf);
				continue;
			}

			//fprintf(tv_out,
			//"#Offset\n%f\n#Slope\n%f\n\n",calib_offsets[i],calib_slopes[i]);
			if(i < 8) { // For LaBr
				fprintf(tv_out,"%.11f\n%.11f\n%.11f\n%.11f\n",
				    calib_offsets[i],calib_slopes[i],calib_quad[i],calib_cubic[i]);
			} else {
				fprintf(tv_out,"%f\n%f\n\n",calib_offsets[i],calib_slopes[i]);
			}
			fclose(tv_out);

		}
	}

	TTimer* t1, *t2;
	TFile* f;
	TTree* ntuple;
	Int_t first;
	TGMainFrame* frame;
	TGTab* tab;
	TRootEmbeddedCanvas* ec[N_FRAMES];
	myTH1F* h[N_HISTOGRAMS];
	myTH2F* h2[N_HISTOGRAMS_2D];
	TGCompositeFrame* cf[N_FRAMES];
	TRandom r;
	TGStatusBar* status;
	const char * filename;

	/* Signals */
	//void StatusBarChanged(); //*SIGNAL*

	/* Slots */
	void RedrawStatusBar() {
		status->Draw();
	}


	/* Branches */
	TBranch* b_labr_h_e[N_LABR];
	TBranch* b_labr_l_e[N_LABR];
	TBranch* b_labrl_t[N_LABR];
	TBranch* b_labr_sis_e[N_LABR];
	TBranch* b_labr_sis_b[N_LABR];
	TBranch* b_labr_s[N_LABR];
	TBranch* b_labrl_s[N_LABR];
	TBranch* b_hpge_e[N_HPGE];
	TBranch* b_hpge_b[N_HPGE];
	TBranch* b_hpge_pileup[N_HPGE];
	TBranch* b_hpgel_t[N_HPGE];
	TBranch* b_hpge_s[N_HPGE];
	TBranch* b_hpge_crm[N_CRM];
	TBranch* b_hpgel_s[N_HPGE];
	TBranch* b_trig_s[N_TRLO];
	TBranch* b_trig_t[N_TRLO];
	TBranch* b_trlo_sc[N_TRLO * 4];
	TBranch* b_tpat;
	TBranch* b_t0;
	TBranch* b_time_lo;
	TBranch* b_time_hi;
	/* Variables */
	UInt_t   labr_h_e[N_LABR];
	UInt_t   labr_l_e[N_LABR];
	UInt_t   labrl_t[N_LABR];
	UInt_t   labr_s[N_LABR];
	UInt_t   labr_s_cur[N_LABR];
	UInt_t   labr_s_prev[N_LABR];
	UInt_t   labr_rate[N_LABR];
	UInt_t   labrl_s[N_LABR];
	UInt_t   labrl_s_cur[N_LABR];
	UInt_t   labrl_s_prev[N_LABR];
	UInt_t   labrl_rate[N_LABR];
	UInt_t   labr_sis_e[N_LABR];
	UInt_t   labr_sis_b[N_LABR];
	UInt_t   hpge_s[N_HPGE];
	UInt_t   hpge_s_cur[N_HPGE];
	UInt_t   hpge_s_prev[N_HPGE];
	UInt_t   hpge_rate[N_HPGE];
	UInt_t   hpgel_s[N_HPGE];
	UInt_t   hpgel_s_cur[N_HPGE];
	UInt_t   hpgel_s_prev[N_HPGE];
	UInt_t   hpge_crm[N_CRM];
	UInt_t   hpge_crm_cur[N_CRM];
	UInt_t   hpge_crm_prev[N_CRM];
	UInt_t   hpge_crm_rate[N_CRM];
	Int_t    hpgel_rate[N_HPGE];
	UInt_t   hpge_e[N_HPGE];
	UInt_t   hpge_e_nobl[N_HPGE];
	UInt_t   hpge_b[N_HPGE];
	UInt_t   hpge_pileup[N_HPGE];
	UInt_t   hpgel_t[N_HPGE];
	UInt_t   tpat;
	UInt_t   t0;
	UInt_t   trig_s[N_TRLO];
	UInt_t   trig_s_prev[N_TRLO];
	UInt_t   trig_rate[N_TRLO];
	UInt_t   trig_t[N_TRLO];
	UInt_t   time_lo;
	UInt_t   time_hi;
	trlo_scaler trlo_sc[N_TRLO];
	trlo_scaler trlo_sc_prev[N_TRLO];
	trlo_scaler64 trlo_sc_sum[N_TRLO];
	float labr_h_e_calib[N_LABR];
	float labr_sis_e_calib[N_LABR];
	float labr_l_e_calib[N_LABR];
	float hpge_e_calib[N_HPGE];
	float calib_slopes[32];
	float calib_offsets[32];
	float calib_cubic[32];
	float calib_quad[32];
	uint64_t start, stop;
	uint64_t ts_first, ts_last, timestamp_begin;
	uint64_t timestamp;
	int ref;
	float t_rf;
	float t_ref;
	int labr_e;
	int only_tv;
	TBox *box_t_vs_rf_labr[N_LABR * 2];
	TBox *box_t_vs_rf_hpge[N_HPGE * 2];

	/* Struct reader */
	FILE	*ucesb_fd;
	ext_data_clnt ucesb_client;
	ext_data_struct_info ucesb_struct_info;
	uint32_t s_eventno;
	uint32_t s_meventno;
	EXT_STR_h101 s_data;
	size_t s_event_size;
	int is_struct_reader;
};

void onKeyPress() {
	if (gPad->GetEvent() == kKeyPress) {
		int key = gPad->GetEventX();
		if (key == 'r') {
			((myTH1F*)(gPad->GetListOfPrimitives()->At(1)))->
			    ResetHist();
		}
	}
}

void TreeReader::struct_reader_init() {
	int ok;
	int ret;

	s_event_size = sizeof(EXT_STR_h101);
	is_struct_reader = 1;

	std::ostringstream command;

	command << UCESB_BIN
		<< " "
		<< filename
		<< " "
		<< "--ntuple="
		<< "RAW"
		<< ",STRUCT,-";
	std::cout << "Calling ucesb with command: " << command.str()
		<< std::endl;
#ifdef USE_PIPE
	ucesb_fd = popen(command.str().c_str(), "r");
#else
	ucesb_fd = stdin;
#endif
	if (nullptr == ucesb_fd) {
		perror("popen()");
		abort();
	}
	ret = ucesb_client.connect(fileno(ucesb_fd));
	if (0 == ret) {
		perror("ext_data_clnt::connect()");
		printf("ucesb: %s\n", ucesb_client.last_error());
		abort();
	}
	EXT_STR_h101_ITEMS_INFO(ok, ucesb_struct_info, 0, EXT_STR_h101, 0);
	if (!ok) {
		perror("ext_data_struct_info_item");
		abort();
	}
	ret = ucesb_client.setup(NULL, 0, &ucesb_struct_info, s_event_size);
	if (0 != ret) {
		perror("ext_data_clnt::setup()");
		printf("ucesb: %s\n", ucesb_client.last_error());
		abort();
	}
}

void TreeReader::struct_reader_close() {
	if (!ucesb_client.close()) {
		perror("ext_data_clnt::close()");
		abort();
	}
	if (nullptr != ucesb_fd) {
		int ret;
		ret = pclose(ucesb_fd);
		if (-1 == ret) {
			perror("plose()");
			abort();
		}
	}
}

void TreeReader::struct_reader_read() {
	int ret;
	ret = ucesb_client.fetch_event(&s_data, s_event_size);
	if (0 == ret) {
		printf("ext_data_clnt::fetch_event: End of input.");
		abort();
	}
	if (-1 == ret) {
		perror("ext_data_clnt::fetch_event()");
		printf("ucesb: %s\n", ucesb_client.last_error());
		abort();
	}
#if 0
	if (!ucesb_client.get_raw_data(&raw, &raw_words)) {
		perror("ext_data_clnt::get_raw_data()");
		abort();
	}
	if (raw) {

	}
#endif
	/*printf("Event data: %10d/%2d (d%10d): %2d\n", s_data.EVENTNO,
	    s_data.MEVENTNO, s_data.EVENTNO - s_eventno, s_data.TRIGGER);*/
}

void TreeReader::struct_reader_get_values() {
	EXT_STR_h101_onion *data = (EXT_STR_h101_onion *)&s_data;

	s_eventno = data->EVENTNO;
	s_meventno = data->MEVENTNO;

	for (int i = 0; i < N_LABR; ++i) {
		labr_sis_e[i] = data->ADC[0].CH[i*4].MAXE[1];
		labr_sis_b[i] = data->ADC[0].CH[i*4].MAXE[0];
		labrl_t[i] = data->LABRL[i].T;
		labrl_s[i] = data->LABRL[i].S;
		labr_s[i] = data->LABRS[i];
	}
	for (int i = 0; i < N_HPGE; ++i) {
		hpge_e[i] = data->ADC[2].CH[i*4].MAXE[1];
		hpge_b[i] = data->ADC[2].CH[i*4].MAXE[0];
		hpgel_t[i] = data->HPGEL[i].T;
		hpgel_s[i] = data->HPGES[i];
	}
	for (int i = 0; i < N_TRIG; ++i) {
		trig_t[i] = data->TRIGT[i];
	}
	for (int i = 0; i < N_CRM; ++i) {
		hpge_crm[i] = data->CRM[i];
	}
	tpat = data->TPAT;
	time_lo = data->TS_LO;
	time_hi = data->TS_HI;
	for (int i = 0; i < N_TRLO; ++i) {
		trlo_sc[i].raw = data->TRLORAW[i];
		trlo_sc[i].bdt = data->TRLOBDT[i];
		trlo_sc[i].adt = data->TRLOADT[i];
		trlo_sc[i].ard = data->TRLOARD[i];
	}
}

void *run_struct_reader_loop(void *t)
{
	return ((TreeReader *)t)->run_struct_reader_loop();
}

void *TreeReader::run_struct_reader_loop()
{
	for(;;) {
		struct_reader_read();
		struct_reader_get_values();
		update_timestamp();

		if (first == 1) {
			ts_first = timestamp;
			timestamp_begin = timestamp;
			first = 0;
		}

		calibrate();
		update_reference_values();
		fillHistograms();
	}
	return 0;
}

pthread_t g_struct_read_thread;

void TreeReader::struct_read_loop(int showProgress) {
	struct timeval now, last_draw_time;

	(void)showProgress;

	first = 1;
	only_tv = 0;

	setupGui();
	styleHistograms();
	setLogScale();
	setupTimers();

	gettimeofday(&last_draw_time, NULL);

	pthread_create(&g_struct_read_thread, nullptr, ::run_struct_reader_loop,
	    this);

	for(;;) {
		gSystem->ProcessEvents();
		gettimeofday(&now, NULL);
		if (now.tv_sec > last_draw_time.tv_sec) {
			redrawHistograms();
			gSystem->ProcessEvents();
			gettimeofday(&now, NULL);
			writeTv();
			last_draw_time = now;
		}
		//every N updateScalerRates();
	}
}

void ghost(const char * filename, const char * options = 0) {
	TApplication a("ghost", 0, nullptr);

	new TreeReader(filename, options);

	a.Run();
}

int main(int argc, char *argv[]) {
	char *_filename;
	char *_options;
	if (argc == 0) {
		abort();
	}
	_filename = argv[1];
	_options = argv[2];
	printf("filename = %s, options = %s\n", _filename, _options);
	ghost(_filename, _options);
	return 0;
}
