#ifndef _GHOST_H
#define _GHOST_H

#define FILENAME "/tmp/test.root"
#define TIMESTAMP_STEP_NS 1
#define DEBUG 0
#define ELOG 0

#define N_LABR 4
#define N_HPGE 4
#define N_TRIG 12
#define N_TRLO 16
#define N_CRM 8

//#define NDET_WIDTH_A1 (float)(100/4)
//#define NDET_WIDTH_A2 (float)(30/4)
//#define NDET_WIDTH_A3 (float)(120/4)

/* HELP:
 *
 * Adding a new Tab:
 *   - Add the tab identifier <TID> to the 'enum FrameName'
 *     The order of this enum reflects the order in the display
 *   - Add the tab name to the array 'f_names'
 *     The order must be the same as in the 'enum FrameName'
 *   - Create the tab by adding to 'setupGui()' function:
 *     createEmbeddedCanvas(<TID>, ny, nx)
 *     <TID> is the tab identifier
 *     ny the number of rows, nx the number of columns
 *
 * Adding a new Histogram (1D):
 *   - Add the histogram identifier <HID> to the 'enum HistogramName'
 *   - Add the histogram data to the 'h_data' array
 *     The order must be the same as in the 'enum HistogramName'
 *   - Create the histogram by adding suitable lines to the
 *     'createHistograms()' function. Just copy and paste from the
 *     existing lines:
 *       h[<HID>] = new myTH1F(h_data[<HID>]);
 *   - Map the histogram ID (HID) to the tab ID (TAB) and pad ID (PID):
 *       canv2hist[<TID>][<PID>] = <HID>;
 *   - Fill the histograms, by adding a function
 *     fill<Det><Property>(), e.g. fillLabrTimes()
 *   - Call this function in the 'fillHistograms()' function
 *
 * Adding a new Histogram (2D):
 *   - Same as adding a 1D histogram, but exchanging some names:
 *     - HistogramName -> HistogramName2
 *     - h_data -> h_data_2d
 *     - h[] -> h2[]
 *     - myTH1F -> myTH2F
 *     - canv2hist -> canv2hist2d
 *
 * Adding new branches:
 *   - Add the branch identifier (BID) to the 'enum BranchName'
 *   - Add the branch name to the 'b_names' array
 *     The order must be the same as in the 'enum BranchName'
 *   - Add variables for the branch and the variable to be filled, e.g.:
 *       TBranch *b_test;
 *       UInt_t test;
 *     These could also be arrays, of course. The type of the variable must
 *     match the type stored in the input Tree.
 *   - Call ntuple->SetBranchAddress(b_names[BID], &test, &b_test) in the
 *     'setBranchAddresses()' function
 *     Now the variable 'test' gets filled with the event data.
 *     
 **/

#if DEBUG > 0
#define PRINTD(...) printf(__VA_ARGS__); fflush(stdout);
#else
#define PRINTD(...)
#endif

/* TPAT */
#define TPAT_LABR_SINGLE_L 1
#define TPAT_HPGE_SINGLE_L 2
#define TPAT_COINCIDENCE_LL 4
#define TPAT_COINCIDENCE_LH 8
#define TPAT_LABR_M 16
#define TPAT_HPGE_M 32
#define TPAT_LABR_SINGLE_H 64
#define TPAT_HPGE_SINGLE_H 128
#define TPAT_LABR_M_LO 256
#define TPAT_NEUTRON_GAMMA 512
#define TPAT_COINCIDENCE_HL 1024

enum TriggerName
{
	TRIG_LABR_OR_L,
	TRIG_LABR_OR_H,
	TRIG_HPGE_OR_L,
	TRIG_HPGE_OR_H,
	TRIG_LABR_M2_L,
	TRIG_LABR_M2_H,
	TRIG_HPGE_M2_L,
	TRIG_HPGE_M2_H,
	TRIG_NDET_OR,
	TRIG_UNUSED,
	TRIG_PADDLE,
	TRIG_RF
};

//#define USED_CONDITION tpat & TPAT_LABR_SINGLE
#define USED_CONDITION 1

/* Colors */
#define RED     "\x1b[31m"
#define GREEN   "\x1b[32m"
#define YELLOW  "\x1b[33m"
#define BLUE    "\x1b[34m"
#define MAGENTA "\x1b[35m"
#define CYAN    "\x1b[36m"
#define BLACK   "\x1b[0m"

/* Branch names */
enum BranchName 
{
/* Trigger logic branches */
	MT, TPAT, TS_LO, TS_HI, TIMESTAMP,
	TRIGT1, TRIGT2, TRIGT3, TRIGT4,
	TRIGT5, TRIGT6, TRIGT7, TRIGT8,
	TRIGT9, TRIGT10, TRIGT11, TRIGT12,
/* LaBr Detector branches */
	LABR1S, LABR2S, LABR3S, LABR4S,
	LABRL1S, LABRL2S, LABRL3S, LABRL4S,
	LABRL1T,LABRL2T,LABRL3T,LABRL4T,
	ADC1CH1MAXE1, ADC1CH5MAXE1, ADC1CH9MAXE1, ADC1CH13MAXE1,
	ADC1CH1MAXE2, ADC1CH5MAXE2, ADC1CH9MAXE2, ADC1CH13MAXE2,
/* HPGe Detector branches */
	HPGE1S, HPGE2S, HPGE3S, HPGE4S,
	HPGEL1S, HPGEL2S, HPGEL3S, HPGEL4S,
	HPGEL1T,HPGEL2T,HPGEL3T,HPGEL4T,
	ADC3CH1MAXE1, ADC3CH5MAXE1, ADC3CH9MAXE1, ADC3CH13MAXE1,
	ADC3CH1MAXE2, ADC3CH5MAXE2, ADC3CH9MAXE2, ADC3CH13MAXE2,
	CRM1, CRM2, CRM3, CRM4, CRM5, CRM6, CRM7, CRM8,
/* Neutron detectors */
	ADC2CH1GATE1, ADC2CH5GATE1, ADC2CH9GATE1, ADC2CH13GATE1,
	ADC2CH2GATE1, ADC2CH6GATE1, ADC2CH10GATE1, ADC2CH14GATE1,
	ADC2CH1GATE2, ADC2CH5GATE2, ADC2CH9GATE2, ADC2CH13GATE2,
	ADC2CH2GATE2, ADC2CH6GATE2, ADC2CH10GATE2, ADC2CH14GATE2,
	ADC2CH1GATE3, ADC2CH5GATE3, ADC2CH9GATE3, ADC2CH13GATE3,
	ADC2CH2GATE3, ADC2CH6GATE3, ADC2CH10GATE3, ADC2CH14GATE3,
	N1E, N2E, N3E, N4E, N5E, N6E, N7E, N8E,
	N1T, N2T, N3T, N4T, N5T, N6T, N7T, N8T,
	N1P, N2P, N3P, N4P, N5P, N6P, N7P, N8P,
	N1S, N2S, N3S, N4S, N5S, N6S, N7S, N8S
};

enum TrloName {
	TRLORAW1,TRLORAW2,TRLORAW3,TRLORAW4,
	TRLORAW5,TRLORAW6,TRLORAW7,TRLORAW8,
	TRLORAW9,TRLORAW10,TRLORAW11,TRLORAW12,
	TRLORAW13,TRLORAW14,TRLORAW15,TRLORAW16,
	TRLOBDT1,TRLOBDT2,TRLOBDT3,TRLOBDT4,
	TRLOBDT5,TRLOBDT6,TRLOBDT7,TRLOBDT8,
	TRLOBDT9,TRLOBDT10,TRLOBDT11,TRLOBDT12,
	TRLOBDT13,TRLOBDT14,TRLOBDT15,TRLOBDT16,
	TRLOADT1,TRLOADT2,TRLOADT3,TRLOADT4,
	TRLOADT5,TRLOADT6,TRLOADT7,TRLOADT8,
	TRLOADT9,TRLOADT10,TRLOADT11,TRLOADT12,
	TRLOADT13,TRLOADT14,TRLOADT15,TRLOADT16,
	TRLOARD1,TRLOARD2,TRLOARD3,TRLOARD4,
	TRLOARD5,TRLOARD6,TRLOARD7,TRLOARD8,
	TRLOARD9,TRLOARD10,TRLOARD11,TRLOARD12,
	TRLOARD13,TRLOARD14,TRLOARD15,TRLOARD16
};

typedef struct trlo_scaler_s {
	uint32_t raw;
	uint32_t bdt;
	uint32_t adt;
	uint32_t ard;
} trlo_scaler;

typedef struct trlo_scaler64_s {
	uint64_t raw;
	uint64_t bdt;
	uint64_t adt;
	uint64_t ard;
} trlo_scaler64;

enum BranchNameCalib
{
	LABR1EC,LABR2EC,LABR3EC,LABR4EC,
	HPGE1EC,HPGE2EC,HPGE3EC,HPGE4EC,
	N1EC,N2EC,N3EC,N4EC,
	N5EC,N6EC,N7EC,N8EC
};

enum BranchNameRecalib
{
	LABR1ECR,LABR2ECR,LABR3ECR,LABR4ECR,
	HPGE1ECR,HPGE2ECR,HPGE3ECR,HPGE4ECR,
	N1ECR,N2ECR,N3ECR,N4ECR,
	N5ECR,N6ECR,N7ECR,N8ECR
};

enum HistogramName
{
	Labr1,  Labr2,  Labr3,  Labr4,
	Labr1_t_raw,  Labr2_t_raw,  Labr3_t_raw,  Labr4_t_raw,
	Labr1_t_RF, Labr2_t_RF, Labr3_t_RF, Labr4_t_RF,
	HPGe1,  HPGe2,  HPGe3,  HPGe4,
	HPGe1_t,  HPGe2_t,  HPGe3_t,  HPGe4_t,
	HPGe1_t_raw,  HPGe2_t_raw,  HPGe3_t_raw,  HPGe4_t_raw,
	HPGe1_t_RF,  HPGe2_t_RF,  HPGe3_t_RF,  HPGe4_t_RF,
	LaBr_Sum, HPGe_Sum,
	Labr1_c,  Labr2_c,  Labr3_c,  Labr4_c,
	HPGe1_c,  HPGe2_c,  HPGe3_c,  HPGe4_c,
	Labr1_t_ref,  Labr2_t_ref,  Labr3_t_ref,  Labr4_t_ref,
	HPGe1_t_ref,  HPGe2_t_ref,  HPGe3_t_ref,  HPGe4_t_ref,
	Labr1condition1, Labr2condition1,
	Labr3condition1, Labr4condition1,
	Labr1condition2, Labr2condition2,
	Labr3condition2, Labr4condition2
};

enum HistogramName2
{
	Coincidence_HPGe_LaBr, Coincidence_LaBr_LaBr,
	HPGe1vsTime, HPGe2vsTime, HPGe3vsTime, HPGe4vsTime,
	HPGe1vsRf, HPGe2vsRf, HPGe3vsRf, HPGe4vsRf,
	HPGe1overTime, HPGe2overTime, HPGe3overTime, HPGe4overTime,
	LaBr1vsTime, LaBr2vsTime, LaBr3vsTime, LaBr4vsTime,
	LaBr1vsRf, LaBr2vsRf, LaBr3vsRf, LaBr4vsRf
};

enum FrameName
{
	RAW_E,
	RAW_T,
	LABR_E,
	LABR_E_COND,
	HPGE_E,
	T_VS_REF,
	T_VS_REF_E,
	T_VS_RF, 
	T_VS_RF_E, 
	SUM_COINC,
	TIMELINES
};

struct Mapping
{
	const char *branch;
	const char *name;
};

#endif
