#!/bin/bash
set -e
set -u

if [ $# -lt 1 ] ; then
  echo Error: No arguments given
  echo Usage: $0 "<stream-file.root> [<cmd>] [<opts>]"
  echo "<cmd> can be either of cal, recal, stat, scl or tv"
  exit
fi

root="root -l -x"
file="$1"

if [ $# -gt 1 ] ; then
  cmd="$2"
else
  cmd=""
fi
if [ $# -gt 2 ] ; then
  opts="$3"
else
  opts=""
fi

CALIB=$(grep -e "^#include \"calib" ghost.C | cut -f2 -d" ")

if [ ! "${cmd}x" == "x" ] ; then
  root="root -l -q -b -x"
  if [[ ! $cmd =~ cal|tv|stat|recal|scl ]] ; then
    echo Usage: $0 "<file.root> [<cmd>]"
    echo "<cmd> can be either of cal, recal, stat, scl or tv"
    exit
  fi
fi

if [[ $file =~ stream ]] ; then
  cmd=""
  echo Running in online spectra mode
  echo ${root} ghost.C+\(\"${file}\",\"${cmd}\"\)
  ${root} ghost.C+\(\"${file}\",\"${cmd}\"\)
fi

if [[ $file =~ run[0-9]{3}.*.root ]] ; then
  calib_to_use="000"
  nr=$(echo $file | grep -Eo "run[0-9]{3}" | grep -Eo "[0-9]{3}")
  echo "Trying to find calibration for run $nr"
  calibs=$(find calib/2014 -name "*.h" | sort -r \
	  | grep -Eo "run[0-9]{3}" | grep -Eo "[0-9]{3}")
  for cal in $calibs; do
    if [ $cal -le $nr ] ; then
      calib_to_use=run${cal}.h
      break;
    fi
  done
  echo "Using calib $calib_to_use"
  mkdir -p /tmp/${USER}_ghost
  cp calib/2014/${calib_to_use} /tmp/${USER}_ghost/ghost_calib.h
#  cp ghost_2014.h /tmp/${USER}_ghost/ghost.h
  cp parse.h /tmp/${USER}_ghost/
  sed ghost.C.2014 -e 's/^\/\/.*$//g' -e 's/calib\/.*$/ghost_calib.h"/' > /tmp/${USER}_ghost/ghost.C
  echo
  echo "${root} /tmp/${USER}_ghost/ghost.C+g(\"${file}\",\"${cmd} ${opts}\")"
  echo 
  ${root} /tmp/${USER}_ghost/ghost.C+g\(\"${file}\",\"${cmd}${opts}\"\)
fi
